# Copyright (c) 2019 Eric Steinberger


import numpy as np
import torch

from PokerRL.eval.lbr import _util
from PokerRL.game.Poker import Poker
from PokerRL.game.PokerRange import PokerRange


class LocalLBRWorker:
    """
    Slave to EvalLBRMaster. Does the LBR computation as described in https://arxiv.org/abs/1612.07547
    """

    def __init__(self, t_prof, chief_handle, eval_agent_cls):
        assert t_prof.n_seats == 2

        self.t_prof = t_prof
        self.lbr_args = t_prof.module_args["lbr"]
        self._eval_env_bldr = _util.get_env_builder_lbr(t_prof=t_prof)
        self.check_to_round = self.lbr_args.lbr_check_to_round

        self.chief_handle = chief_handle

        self.agent = _AgentWrapper(t_prof=t_prof, lbr_args=self.lbr_args, eval_agent_cls=eval_agent_cls)

        # has different raise sizes than agent's env! This needs to be considered when updating the envs after opp acts
        self._env = None
        self.agent_range = PokerRange(env_bldr=self._eval_env_bldr)

        assert self.check_to_round is None or (self.check_to_round in self._eval_env_bldr.rules.ALL_ROUNDS_LIST)

    def run(self, agent_seat_id, n_iterations, mode, stack_size):
        """ returns an estimate of a lower bound of the exploitablity of the agent """

        self.agent.set_mode(mode=mode)
        self.agent.to_stack_size(stack_size)
        self.agent_range.reset()

        self._env = self._eval_env_bldr.get_new_env(is_evaluating=True, stack_size=stack_size)

        if not self.agent.can_compute_mode():
            return None

        if self._eval_env_bldr.env_cls.IS_FIXED_LIMIT_GAME:
            return self._run_limit(agent_seat_id=agent_seat_id, n_iterations=n_iterations)
        else:
            return self._run_no_limit(agent_seat_id=agent_seat_id, n_iterations=n_iterations)

    def update_weights(self, weights_for_eval_agent):
        self.agent.update_weights(weights_for_eval_agent)

    def _reset_episode(self):
        ret = self._env.reset()
        self.agent.reset(deck_state_dict=self._env.cards_state_dict())
        self.agent_range.reset()
        return ret

    def _run_limit(self, agent_seat_id, n_iterations):
        total_lbr_winnings = np.empty(shape=n_iterations, dtype=np.float32)
        lbr_seat_id = 1 - agent_seat_id

        for iteration_id in range(n_iterations):
            if iteration_id % 50 == 0:
                print("LBR hand: ", iteration_id)

            # """""""""""""""""
            # Reset
            # """""""""""""""""
            env_obs, reward, terminal, info = self._reset_episode()

            lbr_hand = self._env.get_hole_cards_of_player(p_id=lbr_seat_id)
            self.agent_range.set_cards_to_zero_prob(cards_2d=lbr_hand)

            # """""""""""""""""
            # Play Episode
            # """""""""""""""""
            while not terminal:
                p_id_acting = self._env.current_player.seat_id

                if self.t_prof.DEBUGGING:
                    assert p_id_acting == self.agent.cpu_agent._internal_env_wrapper.env.current_player.seat_id

                if p_id_acting == lbr_seat_id:
                    # optional feature: check the first N rounds 100% as LBR
                    if (self.check_to_round is not None) and (self._env.current_round < self.check_to_round):
                        action_int = Poker.CHECK_CALL

                    else:
                        _rollout_mngr = _LBRRolloutManager(t_prof=self.t_prof, env_bldr=self._eval_env_bldr,
                                                           env=self._env, lbr_hand_2d=lbr_hand)

                        # illegal: -1, fold: 0, all other: any float
                        _utility = np.full(shape=3, fill_value=-1.0, dtype=np.float32)

                        # ev(s, lbr_a=fold)
                        _utility[Poker.FOLD] = 0.0

                        # ev(s, lbr_a=check_call)
                        _wp = _rollout_mngr.get_lbr_checkdown_equity(
                            agent_range=self.agent_range)  # if check/called down
                        _asked = self._env.seats[agent_seat_id].current_bet - self._env.seats[lbr_seat_id].current_bet
                        _pot_before_action = self._env.get_all_winnable_money()
                        _utility[Poker.CHECK_CALL] = _wp * _pot_before_action - (1 - _wp) * _asked

                        # prepare for raise simulation
                        if Poker.BET_RAISE in self._env.get_legal_actions():
                            _saved_env_state = self._env.state_dict()
                            _saved_agent_env_state = self.agent.env_state_dict()
                            _saved_agent_range_state = self.agent_range.state_dict()

                            # compute ev for raise
                            # _________________________________ simulate LBR play r ____________________________________
                            self._env.step(action=Poker.BET_RAISE)
                            _pot_after_raise = self._env.get_all_winnable_money()

                            self.agent.notify_of_action(p_id_acted=lbr_seat_id, action_he_did=Poker.BET_RAISE)

                            # what agent would do after LBR raises. DOESN'T STEP INTERNAL ENV!
                            _, a_probs_each_hand = self.agent.get_action(step_env=False, need_probs=True)

                            # _______________________________ simulate agent reaction __________________________________
                            # p(agent_fold)
                            _fold_prob = np.sum(self.agent_range.range * a_probs_each_hand[:, Poker.FOLD])

                            # p(not agent_fold | hand)
                            _p_not_fold_per_hand = (1 - a_probs_each_hand[:, Poker.FOLD])

                            # agent_range after not folding
                            self.agent_range.mul_and_norm(_p_not_fold_per_hand)

                            # p(lbr_win | lbr play r -> agent play not fold)
                            _wp_now = _rollout_mngr.get_lbr_checkdown_equity(agent_range=self.agent_range)

                            # ev(state, lbr_a=r)
                            _chips_lbr_puts_in_pot = _pot_after_raise - _pot_before_action
                            _ev_if_fold = _pot_before_action
                            _ev_if_not_fold = (_wp_now * _pot_after_raise) - ((1 - _wp_now) * _chips_lbr_puts_in_pot)
                            _utility[Poker.BET_RAISE] = _fold_prob * _ev_if_fold + (1 - _fold_prob) * _ev_if_not_fold

                            # ________________________________________ reset ___________________________________________
                            self.agent_range.load_state_dict(_saved_agent_range_state)
                            self._env.load_state_dict(_saved_env_state)
                            self.agent.load_env_state_dict(_saved_agent_env_state)

                        # select action with highest approximated EV
                        action_int = np.argmax(_utility)

                    # ________________________________________ notify agent ____________________________________________
                    self.agent.notify_of_action(p_id_acted=lbr_seat_id, action_he_did=action_int)

                else:  # agent has to act
                    action_int, a_probs_each_hand = self.agent.get_action(step_env=True, need_probs=True)
                    self.agent_range.update_after_action(action=action_int,
                                                         all_a_probs_for_all_hands=a_probs_each_hand)

                # _____________________________________________ step ___________________________________________________
                old_game_round = self._env.current_round

                env_obs, reward, terminal, info = self._env.step(action=action_int)

                if self._env.current_round != old_game_round:
                    self.agent_range.update_after_new_round(new_round=self._env.current_round,
                                                            board_now_2d=self._env.board)

            total_lbr_winnings[iteration_id] = reward[lbr_seat_id] * self._env.REWARD_SCALAR * self._env.EV_NORMALIZER

        return total_lbr_winnings

    def _run_no_limit(self, agent_seat_id, n_iterations):
        total_lbr_winnings = np.empty(shape=n_iterations, dtype=np.float32)
        lbr_seat_id = 1 - agent_seat_id
        n_lbr_bets = len(self._env.bet_sizes_list_as_frac_of_pot)

        for iteration_id in range(n_iterations):
            if iteration_id % 50 == 0:
                print("LBR hand: ", iteration_id)

            # """""""""""""""""
            # Reset
            # """""""""""""""""
            env_obs, reward, done, info = self._reset_episode()

            lbr_hand = self._env.get_hole_cards_of_player(p_id=lbr_seat_id)
            self.agent_range.set_cards_to_zero_prob(cards_2d=lbr_hand)

            # """""""""""""""""
            # Play Episode
            # """""""""""""""""
            while not done:
                p_id_acting = self._env.current_player.seat_id

                if self.t_prof.DEBUGGING:
                    assert p_id_acting == self.agent.cpu_agent._internal_env_wrapper.env.current_player.seat_id

                if p_id_acting == lbr_seat_id:

                    # optional feature: check the first N rounds 100% as LBR
                    if (self.check_to_round is not None) and (self._env.current_round < self.check_to_round):
                        action_int = Poker.CHECK_CALL

                    else:
                        _rollout_mngr = _LBRRolloutManager(t_prof=self.t_prof, env_bldr=self._eval_env_bldr,
                                                           env=self._env, lbr_hand_2d=lbr_hand)

                        # illegal: -1, fold: 0, all other: any float
                        _utility = np.full(shape=2 + n_lbr_bets, fill_value=-1.0, dtype=np.float32)

                        # ev(s, lbr_a=fold)
                        _utility[Poker.FOLD] = 0.0

                        # ev(s, lbr_a=check_call)
                        _wp = _rollout_mngr.get_lbr_checkdown_equity(agent_range=self.agent_range)
                        _asked = self._env.seats[agent_seat_id].current_bet - self._env.seats[lbr_seat_id].current_bet
                        _pot_before_action = self._env.get_all_winnable_money()
                        _utility[Poker.CHECK_CALL] = _wp * _pot_before_action - (1 - _wp) * _asked

                        # prepare for raise simulation
                        _saved_env_state = self._env.state_dict()
                        _saved_agent_env_state = self.agent.env_state_dict()
                        _saved_agent_range_state = self.agent_range.state_dict()
                        _legal_raises = self._env.get_legal_actions()
                        for a in [Poker.FOLD, Poker.CHECK_CALL]:
                            if a in _legal_raises:
                                _legal_raises.remove(a)

                        # compute ev for all raise sizes LBR can choose from
                        for r in _legal_raises:
                            raise_frac = self._env.bet_sizes_list_as_frac_of_pot[r - 2]

                            # _________________________________ simulate LBR play r ____________________________________
                            self._env.step(action=r)
                            _pot_after_raise = self._env.get_all_winnable_money()

                            self.agent.notify_of_raise_frac_action(p_id_acted=lbr_seat_id, frac=raise_frac)

                            if self.t_prof.DEBUGGING:
                                assert agent_seat_id == self.agent.cpu_agent._internal_env_wrapper.env.current_player.seat_id

                            # what agent would do after LBR raises. DOESN'T STEP INTERNAL ENV!
                            a_probs_each_hand = self.agent.get_a_probs_for_each_hand()

                            # _______________________________ simulate agent reaction __________________________________
                            # p(agent_fold)
                            _fold_prob = np.sum(self.agent_range.range * a_probs_each_hand[:, Poker.FOLD])

                            # p(not agent_fold | hand)
                            _p_not_fold_per_hand = (1 - a_probs_each_hand[:, Poker.FOLD])

                            # agent_range after not folding
                            self.agent_range.mul_and_norm(_p_not_fold_per_hand)

                            # p(lbr_win | lbr play r -> agent play not fold)
                            _wp_now = _rollout_mngr.get_lbr_checkdown_equity(agent_range=self.agent_range)

                            # ev(state, lbr_a=r)
                            _chips_lbr_puts_in_pot = _pot_after_raise - _pot_before_action
                            _ev_if_fold = _pot_before_action
                            _ev_if_not_fold = (_wp_now * _pot_after_raise) - ((1 - _wp_now) * _chips_lbr_puts_in_pot)
                            _utility[r] = _fold_prob * _ev_if_fold + (1 - _fold_prob) * _ev_if_not_fold

                            # ________________________________________ reset ___________________________________________
                            self.agent_range.load_state_dict(_saved_agent_range_state)
                            self._env.load_state_dict(_saved_env_state)
                            self.agent.load_env_state_dict(_saved_agent_env_state)

                        # select action with highest approximated EV
                        action_int = np.argmax(_utility)

                    # ________________________________________ notify agent ____________________________________________
                    if action_int >= 2:
                        raise_frac = self._env.bet_sizes_list_as_frac_of_pot[action_int - 2]
                        self.agent.notify_of_raise_frac_action(p_id_acted=lbr_seat_id,
                                                               frac=raise_frac)
                    else:
                        self.agent.notify_of_action(p_id_acted=lbr_seat_id,
                                                    action_he_did=action_int)

                else:  # agent has to act
                    if self.t_prof.DEBUGGING:
                        assert p_id_acting == self.agent.cpu_agent._internal_env_wrapper.env.current_player.seat_id

                    action_int, a_probs_each_hand = self.agent.get_action(step_env=True, need_probs=True)

                    self.agent_range.update_after_action(action=action_int,
                                                         all_a_probs_for_all_hands=a_probs_each_hand)
                    if action_int >= 2:
                        # querying what the bet size is in the agent's env_args (these might differ from LBR's!).
                        raise_frac = \
                            self.agent.cpu_agent.env_bldr.env_args.bet_sizes_list_as_frac_of_pot[action_int - 2]

                # _____________________________________________ step ___________________________________________________
                old_game_round = self._env.current_round

                if action_int >= 2:  # step with fraction because agent and LBR have different raise sizes
                    env_obs, reward, done, info = self._env.step_raise_pot_frac(pot_frac=raise_frac)
                else:
                    env_obs, reward, done, info = self._env.step(action=action_int)

                if self._env.current_round != old_game_round:
                    self.agent_range.update_after_new_round(new_round=self._env.current_round,
                                                            board_now_2d=self._env.board)

            total_lbr_winnings[iteration_id] = reward[lbr_seat_id] * self._env.REWARD_SCALAR * self._env.EV_NORMALIZER

        return total_lbr_winnings


class _AgentWrapper:

    def __init__(self, t_prof, lbr_args, eval_agent_cls):
        self.USE_GPU = (t_prof.HAVE_GPU and lbr_args.use_gpu_for_batch_eval and torch.cuda.is_available())

        self.cpu_agent = eval_agent_cls(t_prof=t_prof, device=torch.device("cpu"))

        if self.USE_GPU:
            self.gpu_agent = eval_agent_cls(t_prof=t_prof, device=torch.device("cuda:0"))
            self.gpu_agent.set_env_wrapper(self.cpu_agent.get_env_wrapper())

    # ________________________________ intensive tasks on GPU. Single samples on CPU ___________________________________
    def get_action(self, step_env, need_probs):
        if need_probs and self.USE_GPU:  # if need_probs is True, strategy for all hands is computed --> batch
            return self.gpu_agent.get_action(step_env=step_env, need_probs=need_probs)
        else:
            return self.cpu_agent.get_action(step_env=step_env, need_probs=need_probs)

    def get_a_probs_for_each_hand(self):
        if self.USE_GPU:
            return self.gpu_agent.get_a_probs_for_each_hand()
        else:
            return self.cpu_agent.get_a_probs_for_each_hand()

    # _________________________________________________ just wrapped ___________________________________________________
    def get_mode(self):
        if self.USE_GPU:
            return self.gpu_agent.get_mode()
        return self.cpu_agent.get_mode()

    def set_mode(self, mode):
        self.cpu_agent.set_mode(mode)
        if self.USE_GPU:
            self.gpu_agent.set_mode(mode)

    def to_stack_size(self, stack_size):
        # Automatically sets gpu agent, if applicable, because they share an environment.
        self.cpu_agent.set_stack_size(stack_size=stack_size)

    def can_compute_mode(self):
        return self.cpu_agent.can_compute_mode()

    def update_weights(self, w):
        self.cpu_agent.update_weights(w)
        if self.USE_GPU:
            self.gpu_agent.update_weights(w)

    def reset(self, deck_state_dict):
        self.cpu_agent.reset(deck_state_dict=deck_state_dict)
        if self.USE_GPU:
            self.gpu_agent.reset(deck_state_dict=deck_state_dict)

    def notify_of_action(self, p_id_acted, action_he_did):
        # GPU uses same env. must not do on both!
        self.cpu_agent.notify_of_action(p_id_acted=p_id_acted, action_he_did=action_he_did)

    def notify_of_raise_frac_action(self, p_id_acted, frac):
        self.cpu_agent.notify_of_raise_frac_action(p_id_acted=p_id_acted, frac=frac)

    def env_state_dict(self):
        return self.cpu_agent.env_state_dict()

    def load_env_state_dict(self, state_dict):
        self.cpu_agent.load_env_state_dict(state_dict)


class _LBRRolloutManager:

    def __init__(self, t_prof, env_bldr, env, lbr_hand_2d):
        self.t_prof = t_prof
        self.env_bldr = env_bldr

        self._bigger_idxs = []
        self._equal_idxs = []

        self._env = env

        self._lbr_hand_1d = self.env_bldr.lut_holder.get_1d_cards(cards_2d=lbr_hand_2d)
        self._lbr_hand_range_idx = self.env_bldr.lut_holder.get_range_idx_from_hole_cards(hole_cards_2d=lbr_hand_2d)

        self._board_2d = np.copy(self._env.board)  # still has not-dealt cards
        self._board_1d = self.env_bldr.lut_holder.get_1d_cards(self._board_2d)  # still has not-dealt cards

        self._cards_dealt = np.array([c for c in self._board_1d if c != Poker.CARD_NOT_DEALT_TOKEN_1D])
        self._possible_cards = np.arange(self.env_bldr.rules.N_CARDS_IN_DECK, dtype=np.int32)
        self._possible_cards = np.delete(self._possible_cards, np.concatenate((self._cards_dealt, self._lbr_hand_1d)))

        self._n_cards_to_deal = env_bldr.lut_holder.DICT_LUT_N_CARDS_OUT[self._env.ALL_ROUNDS_LIST[-1]] \
                                - env_bldr.lut_holder.DICT_LUT_N_CARDS_OUT[self._env.current_round]

        # recursion
        self._build_eq_vecs(board_1d=np.copy(self._board_1d),
                            n_cards_to_deal=self._n_cards_to_deal,
                            possible_cards_1d=np.copy(self._possible_cards))

    def _build_eq_vecs(self, board_1d, n_cards_to_deal, possible_cards_1d):
        if n_cards_to_deal > 0:
            """
            only considers cards in possible_cards_1d (sorted arr of ints). For the next street all cards (1d)
            smaller than the one dealt now are NOT CONSIDERED. This way we guarantee to deal each board only once.
            """
            for c in range(possible_cards_1d.shape[0] - (n_cards_to_deal - 1)):
                _possible_cards_1d_next = possible_cards_1d[c + 1:]
                board_1d[-n_cards_to_deal] = possible_cards_1d[c]
                self._build_eq_vecs(board_1d=board_1d,
                                    n_cards_to_deal=n_cards_to_deal - 1,
                                    possible_cards_1d=_possible_cards_1d_next)

        else:  # if we get here, the last round was already dealt. now evaluate.
            handranks = self._env.get_hand_rank_all_hands_on_given_boards(
                boards_1d=board_1d.reshape(1, board_1d.shape[0]), lut_holder=self.env_bldr.lut_holder)[0]

            lbr_hand_rank = handranks[self._lbr_hand_range_idx]
            self._bigger_idxs.append(np.argwhere(handranks < lbr_hand_rank))
            self._equal_idxs.append(np.argwhere(handranks == lbr_hand_rank))

    def get_lbr_checkdown_equity(self, agent_range):
        agent_range_start_sate_dict = agent_range.state_dict()

        # adjust card_probs based on agent's range
        agent_card_probs = agent_range.get_card_probs()
        card_probs = np.subtract(1, agent_card_probs)
        card_probs[self._lbr_hand_1d] = 0.0

        if self.t_prof.DEBUGGING:
            assert np.all(np.less(agent_card_probs[self._lbr_hand_1d], 0.0001))
            assert np.allclose(np.sum(agent_card_probs), self.env_bldr.rules.N_HOLE_CARDS, atol=0.0001)
        cards_dealt = np.array([c for c in self._board_1d if c != Poker.CARD_NOT_DEALT_TOKEN_1D])
        if cards_dealt.shape[0] > 0:
            if self.t_prof.DEBUGGING:
                assert np.all(np.less(agent_card_probs[cards_dealt], 0.00001))
            card_probs[cards_dealt] = 0.0

        # This can be division by 0 in some games, but if it is, it implicitly doesn't matter.
        _s = np.sum(card_probs)
        if _s > 0:
            card_probs /= np.sum(card_probs)

        # recursion
        lbr_win_prob = [0.0]
        self._calc_eq(_win_prob=lbr_win_prob,
                      _i=[0],
                      _board_1d=np.copy(self._board_1d),
                      _n_cards_to_deal=self._n_cards_to_deal,
                      _card_probs=card_probs,
                      _possible_cards_1d=np.copy(self._possible_cards),
                      _reach_prob=1.0,
                      _agent_range=agent_range,
                      _agent_range_start_sate_dict=agent_range_start_sate_dict)

        # winprob has to be multiplied by n_c_to_deal_factorial because we only check one permutation - not all.
        n_c_to_deal_factorial = 1
        for m in range(1, self._n_cards_to_deal + 1):
            n_c_to_deal_factorial *= m

        return lbr_win_prob[0] * n_c_to_deal_factorial

    def _calc_eq(self,
                 _win_prob,
                 _i,
                 _board_1d,
                 _n_cards_to_deal,
                 _card_probs,
                 _possible_cards_1d,
                 _reach_prob,
                 _agent_range,
                 _agent_range_start_sate_dict,
                 ):

        if _n_cards_to_deal > 0:
            """
            only considers cards in possible_cards_1d (sorted arr of ints). For the next street all cards (1d)
            smaller than the one dealt now are NOT CONSIDERED. This way we guarantee to deal each board only once.
            """
            for c in range(_possible_cards_1d.shape[0] - (_n_cards_to_deal - 1)):
                _possible_cards_1d_next = _possible_cards_1d[c + 1:]
                _board_1d[-_n_cards_to_deal] = _possible_cards_1d[c]
                _card_probs_next = np.copy(_card_probs)
                _card_probs_next[_possible_cards_1d[c]] = 0.0
                _card_probs_next /= np.sum(_card_probs_next)
                self._calc_eq(_win_prob=_win_prob,
                              _i=_i,
                              _board_1d=_board_1d,
                              _n_cards_to_deal=_n_cards_to_deal - 1,
                              _possible_cards_1d=_possible_cards_1d_next,
                              _card_probs=_card_probs_next,
                              _reach_prob=_reach_prob * _card_probs[_possible_cards_1d[c]],
                              _agent_range=_agent_range,
                              _agent_range_start_sate_dict=_agent_range_start_sate_dict)

        else:  # if we get here, the last round was already dealt. now evaluate.
            _board_2d = self.env_bldr.lut_holder.get_2d_cards(cards_1d=_board_1d)
            _agent_range.set_cards_to_zero_prob(cards_2d=_board_2d)

            if self.t_prof.DEBUGGING:
                assert not np.any(self._env.CARD_NOT_DEALT_TOKEN_1D == _board_1d)

            lbr_equity = np.sum(_agent_range.range[self._bigger_idxs[_i[0]]])
            lbr_equity += np.sum(_agent_range.range[self._equal_idxs[_i[0]]]) / 2.0

            _win_prob[0] += lbr_equity * _reach_prob
            _agent_range.load_state_dict(_agent_range_start_sate_dict)



# Copyright (c) 2019 Eric Steinberger


from PokerRL.rl.MaybeRay import MaybeRay


class HighLevelAlgoBase:
    """
    A HighLevelAlgo should encapsulate the high-level functionality of an algorithm and should use a
    MaybeRay instance to be compatible with distributed as well as local runs.
    """

    def __init__(self, t_prof, la_handles, chief_handle):
        """

        Args:
            t_prof (TrainingProfile)
            la_handles (list)
            chief_handle (class or ray worker handle)

        """
        self._t_prof = t_prof
        self._ray = MaybeRay(runs_distributed=t_prof.DISTRIBUTED, runs_cluster=t_prof.CLUSTER)
        self._la_handles = la_handles
        self._chief_handle = chief_handle



# Copyright (c) 2019 Eric Steinberger


"""
Wraps the local BR master in a ray actor to be placed on any (one) machine in the cluster. The BR computation itself
is not distributed.
"""

import ray
import torch

from PokerRL.eval.br.LocalBRMaster import LocalBRMaster as LocalEvalBRMaster


@ray.remote(num_cpus=1, num_gpus=1 if torch.cuda.is_available() else 0)
class DistBRMaster(LocalEvalBRMaster):

    def __init__(self, t_prof, chief_handle, eval_agent_cls):
        LocalEvalBRMaster.__init__(self, t_prof=t_prof, chief_handle=chief_handle, eval_agent_cls=eval_agent_cls)



# Copyright (c) 2019 Eric Steinberger
# Inspiration of architecture from DeepStack-Leduc (https://github.com/lifrordi/DeepStack-Leduc/tree/master/Source)
import numpy as np

from PokerRL.game.PokerEnvStateDictEnums import EnvDictIdxs
from PokerRL.game.PokerRange import PokerRange
from PokerRL.game._.tree._.nodes import PlayerActionNode, ChanceNode


class StrategyFiller:

    def __init__(self, tree, env_bldr):
        self._tree = tree
        self._env_bldr = env_bldr
        self._chance_filled = False

    def fill_uniform_random(self):

        if not self._chance_filled:
            self._fill_chance_node_strategy(node=self._tree.root)
            self._chance_filled = True

        self._fill_uniform_random(node=self._tree.root)
        self.update_reach_probs()

    def fill_random_random(self):

        if not self._chance_filled:
            self._fill_chance_node_strategy(node=self._tree.root)
            self._chance_filled = True

        self._fill_random_random(node=self._tree.root)
        self.update_reach_probs()

    def fill_with_agent_policy(self, agent):

        if not self._chance_filled:
            self._fill_chance_node_strategy(node=self._tree.root)
            self._chance_filled = True

        self._fill_with_agent_policy(node=self._tree.root, agent=agent)

        self.update_reach_probs()

    def update_reach_probs(self):
        self._update_reach_probs(node=self._tree.root)

    def _fill_uniform_random(self, node):
        if node is not self._tree.root and node.p_id_acted_last is not self._tree.CHANCE_ID:
            assert node.parent.strategy.shape == (self._env_bldr.rules.RANGE_SIZE, len(node.parent.children),)
            assert np.all(np.abs(np.sum(node.parent.strategy, axis=1) - 1) < 0.001)

        if node.is_terminal:
            return

        # Chance acted last
        if isinstance(node, ChanceNode) or (isinstance(node, PlayerActionNode)
                                            and (not node.is_terminal)
                                            and node.p_id_acting_next != self._tree.CHANCE_ID):
            n_actions = len(node.children)
            node.strategy = np.full(shape=(self._env_bldr.rules.RANGE_SIZE, n_actions),
                                    fill_value=1.0 / float(n_actions))

        for c in node.children:
            self._fill_uniform_random(node=c)

    def _fill_random_random(self, node):
        # assert node reach_probs are set
        # assert np.all(np.abs(np.sum(node.reach_probs, axis=1) - 1) < 0.0001)
        if node is not self._tree.root and node.p_id_acted_last is not self._tree.CHANCE_ID:
            assert node.parent.strategy.shape == (self._env_bldr.rules.RANGE_SIZE, len(node.parent.children),)
            assert np.all(np.abs(np.sum(node.parent.strategy, axis=1) - 1) < 0.001)

        if node.is_terminal:
            return

        # Chance acted last
        if isinstance(node, ChanceNode) or (isinstance(node, PlayerActionNode)
                                            and (not node.is_terminal)
                                            and node.p_id_acting_next != self._tree.CHANCE_ID):
            n_actions = len(node.children)
            node.strategy = np.random.random(size=(self._env_bldr.rules.RANGE_SIZE, n_actions))
            node.strategy /= np.expand_dims(np.sum(node.strategy, axis=1), axis=-1)

        for c in node.children:
            self._fill_random_random(node=c)

    def _fill_with_agent_policy(self, node, agent):
        """
        The agent has to know the reach_probs. Therefore he has to go through all of the previous nodes to build
        be able to output a strategy for a given node. Reach_probs are saved directly under node.reach_probs
        """

        if node is not self._tree.root and node.p_id_acted_last is not self._tree.CHANCE_ID:
            assert node.parent.strategy.shape == (self._env_bldr.rules.RANGE_SIZE, len(node.parent.children),)
            assert np.all(np.abs(np.sum(node.parent.strategy, axis=1) - 1) < 0.001)

        if node.is_terminal:
            return

        # _______________________________________ set strategy of node _________________________________________________
        # Chance acted last or regular action after action node
        if isinstance(node, ChanceNode) or (isinstance(node, PlayerActionNode)
                                            and (not node.is_terminal)
                                            and node.p_id_acting_next != self._tree.CHANCE_ID):
            # fake steps the agent's internal env to the current state
            agent.set_to_public_tree_node_state(node=node)

            assert node.p_id_acting_next == agent._internal_env_wrapper.env.current_player.seat_id, node.p_id_acting_next

            agent_strat = agent.get_a_probs_for_each_hand()
            node.strategy = agent_strat[:, node.allowed_actions]

        # ___________ recursion
        for c in node.children:
            self._fill_with_agent_policy(node=c, agent=agent)

    def _update_reach_probs(self, node):
        if node is not self._tree.root:
            assert node.parent.strategy.shape == (self._env_bldr.rules.RANGE_SIZE, len(node.parent.children),)

        if node.is_terminal:
            return

        # Chance acted last
        if isinstance(node, ChanceNode) or (isinstance(node, PlayerActionNode)
                                            and (not node.is_terminal)
                                            and node.p_id_acting_next != self._tree.CHANCE_ID):
            for c in node.children:
                c.reach_probs = np.copy(node.reach_probs)

                # since arbitrary actions might be illegal in a state, we have to filter which indices to use
                a_idx = node.allowed_actions.index(c.action)
                c.reach_probs[node.p_id_acting_next] = node.strategy[:, a_idx] * node.reach_probs[node.p_id_acting_next]

        # new round gets rolled out now
        elif node.p_id_acting_next == self._tree.CHANCE_ID:
            for c in range(len(node.children)):
                child = node.children[c]
                child.reach_probs = node.reach_probs * node.strategy[:, c]

        else:
            raise TypeError(node)

        for c in node.children:
            self._update_reach_probs(node=c)

    def _fill_chance_node_strategy(self, node):
        assert node.strategy is None
        if node.is_terminal:
            return

        if node.p_id_acting_next == self._tree.CHANCE_ID:
            game_round = node.children[0].env_state[EnvDictIdxs.current_round]
            n_children = len(node.children)
            assert n_children == self._env_bldr.lut_holder.DICT_LUT_N_BOARDS[game_round]

            # chance nodes are uniform random
            node.strategy = np.zeros(shape=(self._env_bldr.rules.RANGE_SIZE, n_children), dtype=np.float32)

            # set strategy for impossible hands to 0
            for c_id in range(n_children):
                mask = PokerRange.get_possible_range_idxs(rules=self._env_bldr.rules,
                                                          lut_holder=self._env_bldr.lut_holder,
                                                          board_2d=node.children[c_id].env_state[EnvDictIdxs.board_2d])
                node.strategy[mask, c_id] = 1.0 / (self._env_bldr.rules.N_CARDS_IN_DECK - 2)

        for c in node.children:
            self._fill_chance_node_strategy(node=c)



# Copyright (c) 2019 Eric Steinberger


import numpy as np
from scipy.special import comb

from PokerRL.game.Poker import Poker
from PokerRL.game.PokerRange import PokerRange
from PokerRL.game._.cpp_wrappers.CppLUT import CppLibHoldemLuts


class _LutGetterBase:

    def __init__(self, rules):
        self.rules = rules

    def get_1d_card_2_2d_card_LUT(self):
        raise NotImplementedError

    def get_2d_card_2_1d_card_LUT(self):
        raise NotImplementedError

    def get_idx_2_hole_card_LUT(self):
        raise NotImplementedError

    def get_hole_card_2_idx_LUT(self):
        raise NotImplementedError

    def get_card_in_what_range_idxs_LUT(self):
        raise NotImplementedError

    def get_range_idx_to_private_obs_LUT(self):
        range_idx_to_hc_lut = self.get_idx_2_hole_card_LUT()
        hc_1d_to_2d_lut = self.get_1d_card_2_2d_card_LUT()

        D = self.rules.N_SUITS + self.rules.N_RANKS

        lut = np.zeros(shape=(self.rules.RANGE_SIZE, D * self.rules.N_HOLE_CARDS), dtype=np.float32)

        for range_idx in range(self.rules.RANGE_SIZE):
            priv_o = np.zeros(shape=self.rules.N_HOLE_CARDS * D, dtype=np.float32)

            for c_id in range(self.rules.N_HOLE_CARDS):
                card = hc_1d_to_2d_lut[range_idx_to_hc_lut[range_idx, c_id]]
                priv_o[D * c_id + card[0]] = 1

                # If the suit doesn't matter, it is not included with the observation.
                if self.rules.SUITS_MATTER:
                    priv_o[D * c_id + self.rules.N_RANKS + card[1]] = 1

            lut[range_idx] = priv_o

        return lut

    def get_n_boards_LUT(self):
        _c = self.get_n_cards_dealt_in_transition_to_LUT()
        return {
            r: comb(N=self.rules.N_RANKS * self.rules.N_SUITS, k=_c[r], exact=True, repetition=False)
            for r in self.rules.ALL_ROUNDS_LIST
        }

    def get_n_cards_out_at_LUT(self):
        return {
            Poker.PREFLOP: 0,
            Poker.FLOP: self.rules.N_FLOP_CARDS,
            Poker.TURN: self.rules.N_FLOP_CARDS + self.rules.N_TURN_CARDS,
            Poker.RIVER: self.rules.N_FLOP_CARDS + self.rules.N_TURN_CARDS + self.rules.N_RIVER_CARDS,
        }

    def get_n_cards_dealt_in_transition_to_LUT(self):
        return {
            Poker.PREFLOP: 0,
            Poker.FLOP: self.rules.N_FLOP_CARDS,
            Poker.TURN: self.rules.N_TURN_CARDS,
            Poker.RIVER: self.rules.N_RIVER_CARDS,
        }

    def get_n_board_branches_LUT(self):
        _N_CARDS_DEALT_IN_TRANSITION_TO_LUT = self.get_n_cards_dealt_in_transition_to_LUT()
        _N_CARDS_OUT_AT = self.get_n_cards_out_at_LUT()
        lut = {
            Poker.PREFLOP: 0
        }
        for r in [_r for _r in self.rules.ALL_ROUNDS_LIST if _r != Poker.PREFLOP]:
            nc = self.rules.N_CARDS_IN_DECK \
                 - _N_CARDS_OUT_AT[self.rules.ROUND_BEFORE[r]] \
                 - self.rules.N_HOLE_CARDS

            # get_range_size is actually a general combinatorial function that we can also use here
            lut[r] = PokerRange.get_range_size(n_hole_cards=_N_CARDS_DEALT_IN_TRANSITION_TO_LUT[r],
                                               n_cards_in_deck=nc)
        return lut


class _LutGetterHoldem(_LutGetterBase):

    def __init__(self, env_cls):
        super().__init__(rules=env_cls.RULES)
        self.cpp_backend = CppLibHoldemLuts(n_boards_lut=self.get_n_boards_LUT(),
                                            n_cards_out_lut=self.get_n_cards_out_at_LUT())

    def get_1d_card_2_2d_card_LUT(self):
        lut = np.full(shape=(self.rules.N_CARDS_IN_DECK, 2), fill_value=-2, dtype=np.int8)
        for c in range(self.rules.N_CARDS_IN_DECK):
            lut[c] = self.cpp_backend.get_2d_card(c)
        return lut

    def get_2d_card_2_1d_card_LUT(self):
        lut = np.full(shape=(self.rules.N_RANKS, self.rules.N_SUITS), fill_value=-2, dtype=np.int8)
        for r in range(self.rules.N_RANKS):
            for s in range(self.rules.N_SUITS):
                lut[r, s] = self.cpp_backend.get_1d_card(card_2d=np.array([r, s], dtype=np.int8))
        return lut

    def get_idx_2_hole_card_LUT(self):
        return self.cpp_backend.get_idx_2_hole_card_lut()

    def get_hole_card_2_idx_LUT(self):
        return self.cpp_backend.get_hole_card_2_idx_lut()

    def get_card_in_what_range_idxs_LUT(self):
        lut = np.full(shape=(self.rules.N_CARDS_IN_DECK, self.rules.N_CARDS_IN_DECK - 1), fill_value=-2,
                      dtype=np.int32)

        _idx2hc_lut = self.get_idx_2_hole_card_LUT()
        for c in range(self.rules.N_CARDS_IN_DECK):
            n = 0
            for range_idx in range(self.rules.RANGE_SIZE):
                if c in _idx2hc_lut[range_idx]:
                    lut[c, n] = range_idx
                    n += 1

        assert not np.any(lut == -2)
        return lut


class _LutGetterLeduc(_LutGetterBase):

    def __init__(self, env_cls):
        super().__init__(rules=env_cls.RULES)

    def get_1d_card_2_2d_card_LUT(self):
        lut = np.full(shape=(self.rules.N_CARDS_IN_DECK, 2), fill_value=-2, dtype=np.int8)
        for c in range(self.rules.N_CARDS_IN_DECK):
            lut[c] = self._get_2d_card(c)
        return lut

    def get_2d_card_2_1d_card_LUT(self):
        lut = np.full(shape=(self.rules.N_RANKS, self.rules.N_SUITS),
                      fill_value=-2, dtype=np.int8)
        for r in range(self.rules.N_RANKS):
            for s in range(self.rules.N_SUITS):
                lut[r, s] = self._get_1d_card(card_2d=np.array([r, s], dtype=np.int8))
        return lut

    def get_idx_2_hole_card_LUT(self):
        # int between 0 and n_cards * (n_cards-1) inclusive --> [c1]
        return np.expand_dims(np.arange(self.rules.N_CARDS_IN_DECK), axis=1)

    def get_hole_card_2_idx_LUT(self):
        # [c1] --> int between 0 and n_cards * (n_cards-1) inclusive
        return np.expand_dims(np.arange(self.rules.N_CARDS_IN_DECK), axis=1)

    def get_card_in_what_range_idxs_LUT(self):
        return np.arange(self.rules.RANGE_SIZE).reshape(-1, 1)  # 1-card games are easy

    def _get_1d_card(self, card_2d):
        """
        Args:
            card_2d (np.ndarray):    array of 2 int8s. [rank, suit]

        Returns:
            int8: 1d representation of card_2d
        """
        return card_2d[0] * self.rules.N_SUITS + card_2d[1]

    def _get_2d_card(self, card_1d):
        """
        Args:
            card_1d (int):

        Returns:
            np.ndarray(shape=2, dtype=np.int8): 2d representation of card_1d
        """
        card_2d = np.empty(shape=2, dtype=np.int8)
        card_2d[0] = card_1d // self.rules.N_SUITS
        card_2d[1] = card_1d % self.rules.N_SUITS
        return card_2d


class _LutHolderBase:
    """ abstract """

    def __init__(self, lut_getter):
        self._lut_getter = lut_getter

        # lut[i, 0] --> rank; ut[i, 1] --> suit
        self.LUT_1DCARD_2_2DCARD = self._lut_getter.get_1d_card_2_2d_card_LUT()
        # lut[rank, suit] --> int
        self.LUT_2DCARD_2_1DCARD = self._lut_getter.get_2d_card_2_1d_card_LUT()
        # lut[range_idx] -> array of size   n_hole_cards * (n_suits + n_ranks)
        self.LUT_RANGE_IDX_TO_PRIVATE_OBS = self._lut_getter.get_range_idx_to_private_obs_LUT()

        self.LUT_IDX_2_HOLE_CARDS = self._lut_getter.get_idx_2_hole_card_LUT()
        self.LUT_HOLE_CARDS_2_IDX = self._lut_getter.get_hole_card_2_idx_LUT()

        # [c] --> list of all range idxs that contain this card.
        self.LUT_CARD_IN_WHAT_RANGE_IDXS = self._lut_getter.get_card_in_what_range_idxs_LUT()

        # [round] -> number of possible public boards in that round
        self.DICT_LUT_N_BOARDS = self._lut_getter.get_n_boards_LUT()

        # [round] -> number of cards that have been dealt until (including) the round
        self.DICT_LUT_N_CARDS_OUT = self._lut_getter.get_n_cards_out_at_LUT()

        # [round] -> number of cards that are dealt in the transition to round
        self.DICT_LUT_CARDS_DEALT_IN_TRANSITION_TO = self._lut_getter.get_n_cards_dealt_in_transition_to_LUT()

        # [round] -> number of possible branches when board is dealt GOING INTO round
        self.DICT_LUT_N_BOARD_BRANCHES = self._lut_getter.get_n_board_branches_LUT()

    def get_1d_card(self, card_2d):
        """
        Args:
            card_2d (np.ndarray):    array of 2 int8s. [rank, suit]

        Returns:
            int8: 1d representation of card_2d
        """

        if card_2d[0] == Poker.CARD_NOT_DEALT_TOKEN_1D:
            return Poker.CARD_NOT_DEALT_TOKEN_1D
        return self.LUT_2DCARD_2_1DCARD[card_2d[0], card_2d[1]]

    def get_1d_cards(self, cards_2d):
        """
        Args:
            cards_2d (iterable):   2D array of shape [N, 2]

        Returns:
            1D array of ints, representing the cards in 1D format
        """
        if len(cards_2d.shape) == 0 or cards_2d.shape[0] == 0:
            return np.array([], dtype=np.int8)

        aa = np.copy(cards_2d)
        aa[aa == Poker.CARD_NOT_DEALT_TOKEN_1D] = 0  # for robustness against not-dealt tokens
        return np.where(cards_2d[:, 0] == Poker.CARD_NOT_DEALT_TOKEN_1D,  # not dealt
                        Poker.CARD_NOT_DEALT_TOKEN_1D,
                        self.LUT_2DCARD_2_1DCARD[aa[:, 0], aa[:, 1]])

    def get_2d_cards(self, cards_1d):
        """
        Args:
            cards_1d (iterable):   list or array of ints. NOT A SINGLE INT!

        Returns:
            2D array of ints representing the cards in 2D format
        """

        if len(cards_1d.shape) == 0 or cards_1d.shape[0] == 0:
            return np.array([], dtype=np.int8)

        aa = np.copy(cards_1d)
        aa[aa == Poker.CARD_NOT_DEALT_TOKEN_1D] = 0  # for robustness against not-dealt tokens
        cards_2d = np.copy(self.LUT_1DCARD_2_2DCARD[aa]).reshape(-1, 2)
        cards_2d[np.where(cards_1d == Poker.CARD_NOT_DEALT_TOKEN_1D)] = Poker.CARD_NOT_DEALT_TOKEN_2D.reshape(2)
        return cards_2d

    def get_range_idx_from_hole_cards(self, hole_cards_2d):
        raise NotImplementedError

    def get_2d_hole_cards_from_range_idx(self, range_idx):
        raise NotImplementedError

    def get_1d_hole_cards_from_range_idx(self, range_idx):
        raise NotImplementedError


class LutHolderLeduc(_LutHolderBase):
    """
    Don't use LUTs from outside this class. use the functions instad!
    """

    def __init__(self, env_cls):
        super().__init__(lut_getter=_LutGetterLeduc(env_cls=env_cls))

    def get_range_idx_from_hole_cards(self, hole_cards_2d):
        c1 = self.get_1d_cards(hole_cards_2d)[0]
        return self.LUT_HOLE_CARDS_2_IDX[c1, 0]

    def get_2d_hole_cards_from_range_idx(self, range_idx):
        c1 = self.LUT_IDX_2_HOLE_CARDS[range_idx, 0]
        return np.array([self.LUT_1DCARD_2_2DCARD[c1]], dtype=np.int8)

    def get_1d_hole_cards_from_range_idx(self, range_idx):
        return np.copy(self.LUT_IDX_2_HOLE_CARDS[range_idx])


class LutHolderHoldem(_LutHolderBase):

    def __init__(self, env_cls):
        super().__init__(lut_getter=_LutGetterHoldem(env_cls=env_cls))

    def get_range_idx_from_hole_cards(self, hole_cards_2d):
        _c1 = self.LUT_2DCARD_2_1DCARD[hole_cards_2d[0, 0]][hole_cards_2d[0, 1]]
        _c2 = self.LUT_2DCARD_2_1DCARD[hole_cards_2d[1, 0]][hole_cards_2d[1, 1]]

        # c1 can never equal c2
        c1 = min(_c1, _c2)
        c2 = max(_c1, _c2)

        return self.LUT_HOLE_CARDS_2_IDX[c1, c2]

    def get_2d_hole_cards_from_range_idx(self, range_idx):
        c1 = self.LUT_IDX_2_HOLE_CARDS[range_idx, 0]
        c2 = self.LUT_IDX_2_HOLE_CARDS[range_idx, 1]

        return np.array([self.LUT_1DCARD_2_2DCARD[c1], self.LUT_1DCARD_2_2DCARD[c2]], dtype=np.int8)

    def get_1d_hole_cards_from_range_idx(self, range_idx):
        return np.copy(self.LUT_IDX_2_HOLE_CARDS[range_idx])



# Copyright (c) 2019 Eric Steinberger


import numpy as np
import torch

from PokerRL.rl.buffers._circular_base import CircularBufferBase


class CircularBufferRNN(CircularBufferBase):
    """
    self.games stores references to Game subclass objects. One Game instance might be referenced multiple times,
    depending on the number of steps that it contains. This is to keep equally likely sampling.
    """

    def __init__(self, env_bldr, max_size):
        super().__init__(env_bldr=env_bldr, max_size=max_size)

        self._games = None
        self.reset()

    @property
    def max_size(self):
        return self._max_size

    @property
    def size(self):
        return self._size

    def add_game(self, game):
        for _ in range(game.n_steps_in_game_memory):
            self._games[self._top] = game

            if self._size < self._max_size:
                self._size += 1

            self._top = (self._top + 1) % self._max_size

    def sample(self, device, batch_size):
        """
        Args:
            batch_size (int)
            device (torch.device)

        Returns:
            tuple
        """
        indices = np.random.randint(low=0, high=self._size, size=batch_size)

        samples = [self._games[i].sample() for i in indices]

        batch_legal_action_mask_tp1 = [sample["mask_tp1"] for sample in samples]
        batch_legal_action_mask_tp1 = torch.from_numpy(np.array(batch_legal_action_mask_tp1)).to(device=device)

        batch_legal_action_mask_t = [sample["mask_t"] for sample in samples]
        batch_legal_action_mask_t = torch.from_numpy(np.array(batch_legal_action_mask_t)).to(device=device)

        batch_action_t = [sample["a"] for sample in samples]
        batch_action_t = torch.tensor(batch_action_t, dtype=torch.long, device=device)

        batch_range_idx = [sample["range_idx"] for sample in samples]
        batch_range_idx = torch.from_numpy(np.array(batch_range_idx)).to(dtype=torch.long, device=device)

        batch_reward = [sample["rew"] for sample in samples]
        batch_reward = torch.from_numpy(np.array(batch_reward)).to(dtype=torch.float32, device=device)

        batch_done = [sample["done"] for sample in samples]
        batch_done = torch.tensor(batch_done, dtype=torch.float32, device=device)

        # obs will be further processed into a PackedSequence in the net.
        batch_pub_obs_t = [sample["o_t"] for sample in samples]
        batch_pub_obs_tp1 = [sample["o_tp1"] for sample in samples]

        return batch_pub_obs_t, \
               batch_action_t, \
               batch_range_idx, \
               batch_legal_action_mask_t, \
               batch_reward, \
               batch_pub_obs_tp1, \
               batch_legal_action_mask_tp1, \
               batch_done,

    def state_dict(self):
        return {
            "games": self._games,
            "size": self._size,
            "top": self._top
        }

    def load_state_dict(self, state):
        self._games = state["games"]
        self._size = state["size"]
        self._top = state["top"]

    def reset(self):
        super().reset()
        self._games = np.array([None for _ in range(self._max_size)], dtype=object)



# Copyright (c) 2019 Eric Steinberger


import numpy as np

from PokerRL.cfr._CFRBase import CFRBase as _CFRBase


class CFRPlus(_CFRBase):

    def __init__(self,
                 name,
                 chief_handle,
                 game_cls,
                 agent_bet_set,
                 starting_stack_sizes=None,
                 delay=0, ):
        """
        delay (int):                            Linear Averaging delay of CFR+ (only applicable if ""cfr_plus"" is
                                                True)
        """
        super().__init__(name=name,
                         chief_handle=chief_handle,
                         game_cls=game_cls,
                         starting_stack_sizes=starting_stack_sizes,
                         agent_bet_set=agent_bet_set,
                         algo_name="CFRp_delay" + str(delay)
                         )

        self.delay = delay
        self.reset()

    def _evaluate_avg_strats(self):
        if self._iter_counter > self.delay:
            return super()._evaluate_avg_strats()

    def _regret_formula_after_first_it(self, ev_all_actions, strat_ev, last_regrets):
        return np.maximum(ev_all_actions - strat_ev + last_regrets, 0)

    def _regret_formula_first_it(self, ev_all_actions, strat_ev):
        return np.maximum(ev_all_actions - strat_ev, 0)  # not max of axis; this is like relu

    def _compute_new_strategy(self, p_id):
        for t_idx in range(len(self._trees)):
            def _fill(_node):
                if _node.p_id_acting_next == p_id:
                    N = len(_node.children)

                    _reg = _node.data["regret"]
                    _reg_sum = np.expand_dims(np.sum(_reg, axis=1), axis=1).repeat(N, axis=1)

                    with np.errstate(divide='ignore', invalid='ignore'):
                        _node.strategy = np.where(
                            _reg_sum > 0.0,
                            _reg / _reg_sum,
                            np.full(shape=(self._env_bldrs[t_idx].rules.RANGE_SIZE, N,), fill_value=1.0 / N,
                                    dtype=np.float32)
                        )

                for c in _node.children:
                    _fill(c)

            _fill(self._trees[t_idx].root)

    def _add_strategy_to_average(self, p_id):
        def _fill(_node):
            if _node.p_id_acting_next == p_id:
                if self._iter_counter > self.delay:
                    current_weight = np.sum(np.arange(self.delay + 1, self._iter_counter + 1))
                    new_weight = self._iter_counter - self.delay + 1

                    m_old = current_weight / (current_weight + new_weight)
                    m_new = new_weight / (current_weight + new_weight)
                    _node.data["avg_strat"] = m_old * _node.data["avg_strat"] + m_new * _node.strategy

                    assert np.allclose(np.sum(_node.data["avg_strat"], axis=1), 1, atol=0.0001)

                elif self._iter_counter == self.delay:
                    _node.data["avg_strat"] = np.copy(_node.strategy)

                    assert np.allclose(np.sum(_node.data["avg_strat"], axis=1), 1, atol=0.0001)

            for c in _node.children:
                _fill(c)

        for t_idx in range(len(self._trees)):
            _fill(self._trees[t_idx].root)



# Copyright (c) 2019 Eric Steinberger


import numpy as np

from PokerRL.rl import rl_util
from PokerRL.rl.buffers._circular_base import BRMemorySaverBase


class BRMemorySaverFLAT(BRMemorySaverBase):
    """ Interface for correct BR reward storing """

    def __init__(self, env_bldr, buffer):
        super().__init__(env_bldr=env_bldr, buffer=buffer)
        self._range_idx = None

    def add_terminal(self,
                     reward_p,
                     terminal_obs,
                     ):

        if self._intermediate_memory.is_level_1():
            self._buffer.add_step(pub_obs_t=np.copy(self._intermediate_memory.obs_t),
                                  a_t=self._intermediate_memory.action,
                                  range_idx=self._range_idx,
                                  r_t=reward_p,
                                  legal_action_mask_t=rl_util.get_legal_action_mask_np(
                                      n_actions=self._env_bldr.N_ACTIONS,
                                      legal_actions_list=self._intermediate_memory.legal_actions_list_t),

                                  pub_obs_tp1=np.copy(terminal_obs),
                                  done_tp1=True,
                                  legal_action_mask_tp1=rl_util.get_legal_action_mask_np(
                                      n_actions=self._env_bldr.N_ACTIONS,
                                      legal_actions_list=self._intermediate_memory.legal_actions_list_tp1)
                                  )

        self._intermediate_memory.reset()

    def add_non_terminal_experience(self,
                                    obs_t_before_acted,
                                    a_selected_t,
                                    legal_actions_list_t):

        self._intermediate_memory.add_experience(obs_t=obs_t_before_acted,
                                                 action_t=a_selected_t,
                                                 legal_actions_list_t=legal_actions_list_t)

        if self._intermediate_memory.is_complete():
            self._add_step_to_memory()
            self._intermediate_memory.step()

    def _add_step_to_memory(self):
        self._buffer.add_step(pub_obs_t=np.copy(self._intermediate_memory.obs_t),
                              a_t=self._intermediate_memory.action,
                              range_idx=self._range_idx,
                              r_t=0.0,
                              legal_action_mask_t=rl_util.get_legal_action_mask_np(
                                  n_actions=self._env_bldr.N_ACTIONS,
                                  legal_actions_list=self._intermediate_memory.legal_actions_list_t),

                              pub_obs_tp1=np.copy(self._intermediate_memory.obs_tp1),
                              done_tp1=False,
                              legal_action_mask_tp1=rl_util.get_legal_action_mask_np(
                                  n_actions=self._env_bldr.N_ACTIONS,
                                  legal_actions_list=self._intermediate_memory.legal_actions_list_tp1))

    def reset(self, range_idx):
        """ Call with env reset """
        self._range_idx = range_idx
        self._intermediate_memory.reset()


# Copyright (c) 2019 Eric Steinberger


import ray

from PokerRL.eval.rl_br.workers.la.Local_RLBR_LearnerActor import Local_RLBR_LearnerActor as _LocalLA


@ray.remote(num_cpus=1)
class Dist_RLBR_LearnerActor(_LocalLA):

    def __init__(self, t_prof, chief_handle, eval_agent_cls):
        super().__init__(t_prof=t_prof, chief_handle=chief_handle, eval_agent_cls=eval_agent_cls)



# Copyright (c) 2019 Eric Steinberger


import numpy as np
import torch
import torch.nn as nn


class MainPokerModuleFLAT(nn.Module):
    """
    Feeds parts of the observation through different fc layers before the RNN

    Structure (each branch merge is a concat):

    Table & Player state --> FC -> RE -> FCS -> RE ----------------------------.
    Board Cards ---> FC -> RE --> cat -> FC -> RE -> FCS -> RE -> FC -> RE --> cat --> FC -> RE -> FCS-> RE -> Normalize
    Private Cards -> FC -> RE -'


    where FCS refers to FC+Skip and RE refers to ReLU
    """

    def __init__(self,
                 env_bldr,
                 device,
                 mpm_args,
                 ):
        super().__init__()
        self.args = mpm_args

        self.env_bldr = env_bldr

        self.N_SEATS = self.env_bldr.N_SEATS
        self.device = device

        self.board_start = self.env_bldr.obs_board_idxs[0]
        self.board_stop = self.board_start + len(self.env_bldr.obs_board_idxs)

        self.pub_obs_size = self.env_bldr.pub_obs_size
        self.priv_obs_size = self.env_bldr.priv_obs_size

        self._relu = nn.ReLU(inplace=False)

        if mpm_args.use_pre_layers:
            self._priv_cards = nn.Linear(in_features=self.env_bldr.priv_obs_size,
                                         out_features=mpm_args.other_units)
            self._board_cards = nn.Linear(in_features=self.env_bldr.obs_size_board,
                                          out_features=mpm_args.other_units)

            self.cards_fc_1 = nn.Linear(in_features=2 * mpm_args.other_units, out_features=mpm_args.card_block_units)
            self.cards_fc_2 = nn.Linear(in_features=mpm_args.card_block_units, out_features=mpm_args.card_block_units)
            self.cards_fc_3 = nn.Linear(in_features=mpm_args.card_block_units, out_features=mpm_args.other_units)

            self.hist_and_state_1 = nn.Linear(in_features=self.env_bldr.pub_obs_size - self.env_bldr.obs_size_board,
                                              out_features=mpm_args.other_units)
            self.hist_and_state_2 = nn.Linear(in_features=mpm_args.other_units, out_features=mpm_args.other_units)

            self.final_fc_1 = nn.Linear(in_features=2 * mpm_args.other_units, out_features=mpm_args.other_units)
            self.final_fc_2 = nn.Linear(in_features=mpm_args.other_units, out_features=mpm_args.other_units)

        else:
            self.final_fc_1 = nn.Linear(in_features=self.env_bldr.complete_obs_size, out_features=mpm_args.other_units)
            self.final_fc_2 = nn.Linear(in_features=mpm_args.other_units, out_features=mpm_args.other_units)

        self.lut_range_idx_2_priv_o = torch.from_numpy(self.env_bldr.lut_holder.LUT_RANGE_IDX_TO_PRIVATE_OBS)
        self.lut_range_idx_2_priv_o = self.lut_range_idx_2_priv_o.to(device=self.device, dtype=torch.float32)

        self.to(device)

    @property
    def output_units(self):
        return self.args.other_units

    def forward(self, pub_obses, range_idxs):
        """
        1. do list -> padded
        2. feed through pre-processing fc layers
        3. PackedSequence (sort, pack)
        4. rnn
        5. unpack (unpack re-sort)
        6. cut output to only last entry in sequence

        Args:
            pub_obses (list):                 list of np arrays of shape [np.arr([history_len, n_features]), ...)
            range_idxs (LongTensor):        range_idxs (one for each pub_obs) tensor([2, 421, 58, 912, ...])
        """

        # ____________________________________________ Packed Sequence _____________________________________________
        priv_obses = self.lut_range_idx_2_priv_o[range_idxs]

        if isinstance(pub_obses, list):
            pub_obses = torch.from_numpy(np.array(pub_obses)).to(self.device, torch.float32)

        if self.args.use_pre_layers:
            _board_obs = pub_obses[:, self.board_start:self.board_stop]
            _hist_and_state_obs = torch.cat([
                pub_obses[:, :self.board_start],
                pub_obses[:, self.board_stop:]
            ],
                dim=-1
            )
            y = self._feed_through_pre_layers(board_obs=_board_obs, priv_obs=priv_obses,
                                              hist_and_state_obs=_hist_and_state_obs)

        else:
            y = torch.cat((priv_obses, pub_obses,), dim=-1)

        final = self._relu(self.final_fc_1(y))
        final = self._relu(self.final_fc_2(final) + final)

        # Normalize last layer
        if self.args.normalize:
            final = final - final.mean(dim=-1).unsqueeze(-1)
            final = final / final.std(dim=-1).unsqueeze(-1)

        return final

    def _feed_through_pre_layers(self, priv_obs, board_obs, hist_and_state_obs):

        # """""""""""""""
        # Cards Body
        # """""""""""""""
        _priv_1 = self._relu(self._priv_cards(priv_obs))
        _board_1 = self._relu(self._board_cards(board_obs))

        cards_out = self._relu(self.cards_fc_1(torch.cat([_priv_1, _board_1], dim=-1)))
        cards_out = self._relu(self.cards_fc_2(cards_out) + cards_out)
        cards_out = self.cards_fc_3(cards_out)

        hist_and_state_out = self._relu(self.hist_and_state_1(hist_and_state_obs))
        hist_and_state_out = self.hist_and_state_2(hist_and_state_out) + hist_and_state_out

        return self._relu(torch.cat([cards_out, hist_and_state_out], dim=-1))


class MPMArgsFLAT:

    def __init__(self,
                 use_pre_layers=True,
                 card_block_units=192,
                 other_units=64,
                 normalize=True,
                 ):
        self.use_pre_layers = use_pre_layers
        self.other_units = other_units
        self.card_block_units = card_block_units
        self.normalize = normalize

    def get_mpm_cls(self):
        return MainPokerModuleFLAT



# Copyright (c) 2019 Eric Steinberger


import copy


class EnvWrapperBuilderBase:
    WRAPPER_CLS = NotImplementedError

    def __init__(self,
                 env_cls,
                 env_args,
                 ):
        """
        Args:
            env_cls (PokerEnv subclass):    Any PokerEnv subclass (not an instance)

            env_args:                       An instance of either PokerEnvArgs or DiscretePokerEnv, depending on what
                                            type env_cls is
        """
        self.env_cls = env_cls
        self.rules = env_cls.RULES
        self.env_args = env_args

        self.lut_holder = env_cls.get_lut_holder()

        # Info about the env
        self.N_SEATS = env_args.n_seats
        self.N_ACTIONS = env_args.N_ACTIONS
        self.pub_obs_size = self._get_num_public_observation_features()
        self.priv_obs_size = self._get_num_private_observation_features()
        self.complete_obs_size = self.pub_obs_size + self.priv_obs_size
        self.obs_board_idxs, self.obs_players_idxs, self.obs_table_state_idxs = self._get_obs_parts_idxs()
        self.obs_size_board = len(self.obs_board_idxs)
        self.obs_size_player_info_each = len(self.obs_players_idxs[0])
        self.obs_size_table_state = len(self.obs_table_state_idxs)

    def get_new_env(self, is_evaluating, stack_size=None):
        """
        Args:
            is_evaluating (bool):       if True, no stack-size randomization is applied to the environment.

            stack_size (list):          Optional.
                                        list of n_seats ints. if provided: the environment will use these stack-
                                        sizes as the default starting point. If not provided, the env will use
                                        the standard stack sizes given by the EnvBuilder instance this is called in.

        Returns:
            PokerEnv subclass instance
        """

        args = copy.deepcopy(self.env_args)
        if stack_size is not None:
            assert isinstance(stack_size, list)
            args.starting_stack_sizes_list = copy.deepcopy(stack_size)

        return self.env_cls(env_args=args, lut_holder=self.lut_holder, is_evaluating=is_evaluating)

    def get_new_wrapper(self, is_evaluating, init_from_env=None, stack_size=None):
        """
        Args:
            init_from_env:              Optional.
                                        An instance of the same PokerEnv subclass to be created with this wrapper.
                                        Its deck, etc. is used to initialize the one to the exact same state and
                                        future randomness.

            is_evaluating (bool):       if True, no stack-size randomization is applied to the environment.

            stack_size (int):           Optional.
                                        list of n_seats ints. if provided: the environment will use these stack-
                                        sizes as the default starting point. If not provided, the env will use
                                        the standard stack sizes given by the EnvBuilder instance this is called in.

        Returns:
            EnvWrapper subclass instance
        """
        if init_from_env is None:
            env = self.get_new_env(is_evaluating=is_evaluating, stack_size=stack_size)
        else:
            env = init_from_env
        return self.WRAPPER_CLS(env=env, env_bldr_that_built_me=self)

    def _get_num_public_observation_features(self):
        """ Can be overridden if needed.

        Return the number of features (potentially per timestep) of the wrapped observation.
        This is only run once, so it is ok, if it is not efficient to evaluate this.
        """
        _env = self.env_cls(env_args=self.env_args, lut_holder=self.lut_holder, is_evaluating=True)
        return _env.observation_space.shape[0]

    def _get_num_private_observation_features(self):
        """
        Can be overridden if needed.
        Return the number of features (potentially per timestep) of the vector that will be appended to the public obs
        to resemble the private obs.
        This is only run once, so it is ok, if it is not efficient to evaluate this.
        """
        return (self.rules.N_SUITS + self.rules.N_RANKS) * self.rules.N_HOLE_CARDS

    def _get_obs_parts_idxs(self):
        """
        Override if your wrapper appends something in front of the env's original obs.
        Returns:

        """
        _env = self.env_cls(env_args=self.env_args, lut_holder=self.lut_holder, is_evaluating=True)
        return \
            _env.obs_parts_idxs_dict["board"], \
            _env.obs_parts_idxs_dict["players"], \
            _env.obs_parts_idxs_dict["table_state"]



# Copyright (c) 2019 Eric Steinberger


import numpy as np

from PokerRL.eval._.EvaluatorMasterBase import EvaluatorMasterBase
from PokerRL.rl import rl_util


class LocalHead2HeadMaster(EvaluatorMasterBase):
    """
    Evaluates two modes of an EvalAgent against each other.
    """

    def __init__(self, t_prof, chief_handle, eval_agent_cls):
        super().__init__(t_prof=t_prof, eval_env_bldr=rl_util.get_env_builder(t_prof=t_prof), chief_handle=chief_handle,
                         eval_type="Head2Head_Winnings", log_conf_interval=True)

        self._args = t_prof.module_args["h2h"]
        self._env_bldr = rl_util.get_env_builder(t_prof=t_prof)

        assert self._env_bldr.N_SEATS == 2

        self._eval_agents = [
            eval_agent_cls(t_prof=t_prof)
            for _ in range(self._env_bldr.N_SEATS)
        ]

        self._REFERENCE_AGENT = 0

    def set_modes(self, modes):
        for e, mode in zip(self._eval_agents, modes):
            e.set_mode(mode)

    def update_weights(self):
        w = self.pull_current_strat_from_chief()

        # Assumes that each eval_agent knows what to do with this dict. This should be the case after .set_modes() has
        # been called
        for e in self._eval_agents:
            e.update_weights(w)

    def evaluate(self, iter_nr):
        """ assumes same action space between all eval agents!! """

        if self._is_multi_stack:
            total_of_all_stacks = []
            pm_of_all_stacks = []

        for stack_size_idx, stack_size in enumerate(self._t_prof.eval_stack_sizes):
            for e in self._eval_agents:
                e.set_stack_size(stack_size=stack_size)

            do_it = True
            for e in self._eval_agents:
                if not e.can_compute_mode():
                    do_it = False
                    break

            if do_it:
                mean, d = self._run_eval(stack_size=stack_size)
                self._log_results(iter_nr=iter_nr,
                                  agent_mode=self._eval_agents[self._REFERENCE_AGENT].get_mode(),
                                  stack_size_idx=stack_size_idx,
                                  score=mean, upper_conf95=mean + d, lower_conf95=mean - d)

            if self._is_multi_stack:
                total_of_all_stacks.append(mean)
                pm_of_all_stacks.append(d)

        if self._is_multi_stack:
            _mean = sum(total_of_all_stacks) / float(len(total_of_all_stacks))
            _d = sum(pm_of_all_stacks) / float(len(pm_of_all_stacks))
            self._log_multi_stack(agent_mode="Head2Head",
                                  iter_nr=iter_nr,
                                  score_total=_mean,
                                  lower_conf95=_mean - _d,
                                  upper_conf95=_mean + _d,
                                  )

    def _run_eval(self, stack_size):
        winnings = np.empty(shape=(self._args.n_hands * self._env_bldr.N_SEATS), dtype=np.float32)

        _env = self._eval_env_bldr.get_new_env(is_evaluating=True, stack_size=stack_size)

        for seat_p0 in range(self._env_bldr.N_SEATS):
            seat_p1 = 1 - seat_p0

            for iteration_id in range(self._args.n_hands):

                # """""""""""""""""
                # Reset
                # """""""""""""""""
                _, r_for_all, done, info = _env.reset()
                for e in self._eval_agents:
                    e.reset(deck_state_dict=_env.cards_state_dict())

                # """""""""""""""""
                # Play Episode
                # """""""""""""""""
                while not done:
                    p_id_acting = _env.current_player.seat_id

                    if p_id_acting == seat_p0:
                        action_int, _ = self._eval_agents[self._REFERENCE_AGENT].get_action(step_env=True,
                                                                                            need_probs=False)
                        self._eval_agents[1 - self._REFERENCE_AGENT].notify_of_action(p_id_acted=p_id_acting,
                                                                                      action_he_did=action_int)
                    elif p_id_acting == seat_p1:
                        action_int, _ = self._eval_agents[1 - self._REFERENCE_AGENT].get_action(step_env=True,
                                                                                                need_probs=False)
                        self._eval_agents[self._REFERENCE_AGENT].notify_of_action(p_id_acted=p_id_acting,
                                                                                  action_he_did=action_int)
                    else:
                        raise ValueError("Only HU supported!")

                    _, r_for_all, done, info = _env.step(action_int)

                # """""""""""""""""
                # Add Rews
                # """""""""""""""""
                winnings[iteration_id + (seat_p0 * self._args.n_hands)] = r_for_all[seat_p0] \
                                                                          * _env.REWARD_SCALAR \
                                                                          * _env.EV_NORMALIZER

        return self._get_95confidence(winnings)



from . import PokerEnvStateDictEnums
from . import bet_sets
from .AgentTournament import *
from .InteractiveGame import *
from .Poker import *
from .PokerRange import *
from .games import StandardLeduc, BigLeduc, NoLimitLeduc, DiscretizedNLLeduc, LimitHoldem, NoLimitHoldem, \
    DiscretizedNLHoldem, Flop5Holdem
from .wrappers import *



# Copyright (c) 2019 Eric Steinberger


import unittest
from unittest import TestCase

from PokerRL.game._.rl_env.base._Deck import DeckOfCards


class Test(TestCase):

    def test_build(self):
        d = DeckOfCards()
        assert d.deck_remaining.shape == (d.n_ranks * d.n_suits, 2)

    def test_draw(self):
        for n in range(20):
            deck = DeckOfCards()
            cards = deck.draw(n)
            assert deck.deck_remaining.shape == (deck.n_ranks * deck.n_suits - n, 2)

            for card in cards:
                for _card in deck.deck_remaining:
                    assert not (card[0] == _card[0] and card[1] == _card[1])


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2019 Eric Steinberger


class _PokerEnvArgs:

    def __init__(self,
                 n_seats,
                 starting_stack_sizes_list=None,
                 stack_randomization_range=(0, 0),
                 scale_rewards=True,
                 use_simplified_headsup_obs=True,
                 return_pre_transition_state_in_info=False,
                 *args,
                 **kwargs,
                 ):
        """
        Args:
            n_seats (int):                      The number of players in the game

            starting_stack_sizes_list (list):   An integer for each player, specifying the starting stack size.

            stack_randomization_range (tuple): (min_delta, max_delta). This will be added to the specified stack sizes
                                                Stack sizes are going to be subject to random offsets in the set range
                                                each episode. This only applies in evaluation mode of the environment.
                                                To not use this feature, pass (0,0).

            scale_rewards (bool):               Whether to scale rewards or not

            use_simplified_headsup_obs (bool):  Whether HU (i.e. 1v1) envs should have an obs without game aspects only
                                                relevant in 3+ player games (e.g. side-pots).

            return_pre_transition_state_in_info (bool):
                                                Whether the environment shall return certain additional information
        """
        self.n_seats = n_seats

        if starting_stack_sizes_list is None:
            self.starting_stack_sizes_list = [None for _ in range(n_seats)]
        else:
            self.starting_stack_sizes_list = starting_stack_sizes_list
        self.stack_randomization_range = stack_randomization_range
        self.scale_rewards = scale_rewards
        self.use_simplified_headsup_obs = use_simplified_headsup_obs
        self.RETURN_PRE_TRANSITION_STATE_IN_INFO = return_pre_transition_state_in_info


class NoLimitPokerEnvArgs(_PokerEnvArgs):
    """
    Args to any game that inherits from NoLimitPokerEnv
    """

    def __init__(self,
                 n_seats,
                 starting_stack_sizes_list=None,
                 stack_randomization_range=(0, 0),
                 scale_rewards=True,
                 use_simplified_headsup_obs=True,
                 return_pre_transition_state_in_info=False,
                 *args,
                 **kwargs):
        super().__init__(n_seats=n_seats,
                         starting_stack_sizes_list=starting_stack_sizes_list,
                         stack_randomization_range=stack_randomization_range,
                         scale_rewards=scale_rewards,
                         use_simplified_headsup_obs=use_simplified_headsup_obs,
                         return_pre_transition_state_in_info=return_pre_transition_state_in_info,
                         *args, **kwargs)
        self.N_ACTIONS = 3


class DiscretizedPokerEnvArgs(_PokerEnvArgs):
    """
    Args to any game that inherits from DiscretizedPokerEnv.
    """

    def __init__(self,
                 n_seats,
                 bet_sizes_list_as_frac_of_pot,
                 starting_stack_sizes_list=None,
                 stack_randomization_range=(0, 0),
                 uniform_action_interpolation=False,
                 use_simplified_headsup_obs=True,
                 scale_rewards=True,
                 return_pre_transition_state_in_info=False,
                 *args, **kwargs):
        """
        Args:
            bet_sizes_list_as_frac_of_pot (list):   list of allowed bet sizes in fractions of current pot.
                                                    e.g. [0.1, 0.3, 0.7, 1, 1.5]

            uniform_action_interpolation (bool):    In discrete poker envs, a finite number of raise sizes is defined.
                                                    If, for instance, the bet sizes are [0.1, 0.3, 0.7, 1, 1.5],
                                                    the agent doesn't bet these exact amounts, but uniformly sampled
                                                    amounts between the selected bet size, and the next bigger and
                                                    smaller one, if this argument is set to True.

            for info on all other arguments refer to docs in PokerEnvArgs' init function
        """
        super().__init__(n_seats=n_seats,
                         starting_stack_sizes_list=starting_stack_sizes_list,
                         stack_randomization_range=stack_randomization_range,
                         scale_rewards=scale_rewards,
                         use_simplified_headsup_obs=use_simplified_headsup_obs,
                         return_pre_transition_state_in_info=return_pre_transition_state_in_info,
                         *args, **kwargs)
        self.bet_sizes_list_as_frac_of_pot = bet_sizes_list_as_frac_of_pot
        self.uniform_action_interpolation = uniform_action_interpolation
        self.N_ACTIONS = len(self.bet_sizes_list_as_frac_of_pot) + 2  # +2 is for FOLD and CHECK/CALL.


class LimitPokerEnvArgs(_PokerEnvArgs):
    """
    Args to any game that inherits from LimitPokerEnv
    """

    def __init__(self,
                 n_seats,
                 starting_stack_sizes_list=None,
                 stack_randomization_range=(0, 0),
                 use_simplified_headsup_obs=True,
                 scale_rewards=True,
                 return_pre_transition_state_in_info=False,
                 *args, **kwargs):
        super().__init__(n_seats=n_seats,
                         starting_stack_sizes_list=starting_stack_sizes_list,
                         stack_randomization_range=stack_randomization_range,
                         scale_rewards=scale_rewards,
                         use_simplified_headsup_obs=use_simplified_headsup_obs,
                         return_pre_transition_state_in_info=return_pre_transition_state_in_info,
                         *args, **kwargs)
        self.N_ACTIONS = 3



# Copyright (c) 2019 Eric Steinberger


import ctypes
import os

import numpy as np


class CppWrapper:
    """
    Baseclass to wrap a C++ library using ctypes
    """
    ARR_2D_ARG_TYPE = np.ctypeslib.ndpointer(dtype=np.intp, ndim=1, flags='C')
    CPP_LIB_FILE_ENDING = "dll" if os.name == 'nt' else "so"

    def __init__(self, path_to_dll):
        self._clib = ctypes.cdll.LoadLibrary(path_to_dll)

    @staticmethod
    def np_1d_arr_to_c(np_arr):
        return ctypes.c_void_p(np_arr.ctypes.data)

    @staticmethod
    def np_2d_arr_to_c(np_2d_arr):
        return (np_2d_arr.__array_interface__['data'][0]
                + np.arange(np_2d_arr.shape[0]) * np_2d_arr.strides[0]).astype(np.intp)



# Copyright (c) 2019 Eric Steinberger


import os.path as osp

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('%s/%s' % (osp.dirname(osp.realpath(__file__)), 'requirements.txt')) as f:
    requirements = [line.strip() for line in f]

with open('%s/%s' % (osp.dirname(osp.realpath(__file__)), 'requirements_dist.txt')) as f:
    requirements_dist = [line.strip() for line in f]

setuptools.setup(
    name="PokerRL",
    version="0.0.1",
    author="Eric Steinberger",
    author_email="ericsteinberger.est@gmail.com",
    description="A framework for Reinforcement Learning in Poker.",
    license='MIT',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TinkeringCode/PokerRL",
    install_requires=requirements,
    extras_require={
        'distributed': requirements_dist,
    },
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
)



# Copyright (c) 2019 Eric Steinberger


import copy

import numpy as np

from PokerRL.eval._.EvaluatorMasterBase import EvaluatorMasterBase
from PokerRL.eval.rl_br import _util
from PokerRL.rl.agent_modules.DDQN import DDQN


class LocalRLBRMaster(EvaluatorMasterBase):

    def __init__(self, t_prof, chief_handle, eval_agent_cls):
        super().__init__(t_prof=t_prof,
                         eval_env_bldr=_util.get_env_builder_rlbr(t_prof=t_prof),
                         chief_handle=chief_handle,
                         eval_type="RL-BR",
                         log_conf_interval=True)

        assert self._eval_env_bldr.N_SEATS == 2, "only works for 2 players at the moment"

        self._args = t_prof.module_args["rlbr"]
        self._eval_agent = eval_agent_cls(t_prof=t_prof)

        self._la_handles = None
        self._ps_handle = None

    def set_learner_actors(self, *las):
        self._la_handles = list(las)

    def set_param_server(self, param_server):
        self._ps_handle = param_server

    def update_weights(self):
        w = self.pull_current_strat_from_chief()
        self._eval_agent.update_weights(copy.deepcopy(w))

    def evaluate(self, global_iter_nr):

        for mode in self._t_prof.eval_modes_of_algo:
            for stack_size_idx, stack_size in enumerate(self._t_prof.eval_stack_sizes):
                self._eval_agent.set_mode(mode=mode)
                self._eval_agent.set_stack_size(stack_size=stack_size)
                if self._eval_agent.can_compute_mode():
                    self._retrain(mode=mode, stack_size=stack_size, stack_size_idx=stack_size_idx,
                                  global_iter_nr=global_iter_nr)
                    # """""""""""""""""""
                    # Compute RL-BR
                    # """""""""""""""""""
                    print("Running rollout matches between RL-BR and agent.")
                    ddqn_states = self._ray.get(self._ray.remote(self._ps_handle.get_eval_ddqn_state_dicts))
                    ddqns = [
                        DDQN.inference_version_from_state_dict(state_dict=ddqn_states[p],
                                                               env_bldr=self._eval_env_bldr)
                        for p in range(self._t_prof.n_seats)
                    ]
                    scores = self._compute_rlbr(
                        n_hands_each_seat=self._args.n_hands_each_seat,
                        rlbr_dqn_each_seat=ddqns,
                        rlbr_env_wrapper=self._eval_env_bldr.get_new_wrapper(is_evaluating=True, stack_size=stack_size),
                        opponent=self._eval_agent
                    )

                    mean, d = self._get_95confidence(scores=scores)

                    self._log_results(iter_nr=global_iter_nr, agent_mode=mode, stack_size_idx=stack_size_idx,
                                      score=mean, lower_conf95=mean - d, upper_conf95=mean + d)

    def _retrain(self, mode, stack_size, stack_size_idx, global_iter_nr):
        # """""""""""""""""""
        # Prepare Logging
        # """""""""""""""""""
        if self._t_prof.log_verbose:
            running_rew_exp = self._ray.get(
                self._ray.remote(self._chief_handle.create_experiment,
                                 self._t_prof.name + "_M_" + mode + "_S" + str(stack_size_idx) + "_I" + str(
                                     global_iter_nr) + "Running Rew RL-BR"))
            eps_exp = self._ray.get([
                self._ray.remote(self._chief_handle.create_experiment,
                                 self._t_prof.name + "_M_" + mode + "_S" + str(stack_size_idx) + "_I" + str(
                                     global_iter_nr) + "Epsilon RL-BR P" + str(p))
                for p in range(self._eval_env_bldr.N_SEATS)
            ])

        logging_scores_each_p = [[] for _ in range(self._eval_env_bldr.N_SEATS)]
        logging_eps_per_p = [[] for _ in range(self._eval_env_bldr.N_SEATS)]
        logging_timesteps = []

        # """""""""""""""""""
        # Training
        # """""""""""""""""""
        self._ray.wait([
            self._ray.remote(self._ps_handle.reset, p, global_iter_nr)
            for p in range(self._eval_env_bldr.N_SEATS)
        ])

        for p_training in range(self._eval_env_bldr.N_SEATS):
            print("Training RL-BR seat", p_training, "with agent mode", mode, "and stack size", stack_size_idx)

            # """"""""
            # Reset
            # """"""""
            self._eval_agent.reset()

            self._ray.wait([
                self._ray.remote(la.reset,
                                 p_training, self._eval_agent.state_dict(), stack_size)
                for la in self._la_handles
            ])
            self._update_leaner_actors(update_eps_for_plyrs=[p_training], update_net_for_plyrs=[p_training])

            # """"""""
            # Pre-play
            # """"""""
            self._ray.wait([
                self._ray.remote(la.play, self._args.pretrain_n_games)
                for la in self._la_handles
            ])

            # """"""""
            # Learn
            # """"""""
            SMOOTHING = 200
            accum_score = 0.0
            for training_iter_id in range(self._args.n_iterations):
                self._ray.wait([
                    self._ray.remote(self._ps_handle.update_eps, p_training, training_iter_id)
                ])

                self._update_leaner_actors(update_eps_for_plyrs=[p_training], update_net_for_plyrs=[p_training])

                # Play
                scores_all_las = self._ray.get([
                    self._ray.remote(la.play, self._args.play_n_games_per_iter)
                    for la in self._la_handles
                ])

                accum_score += sum(scores_all_las) / self._args.n_las

                # Get Gradients
                grads_from_all_las = self._get_gradients(p_id=p_training)

                # Applying gradients
                self._ray.wait([
                    self._ray.remote(self._ps_handle.apply_grads,
                                     p_training, grads_from_all_las)
                ])

                # Update weights on all las
                self._update_leaner_actors(update_net_for_plyrs=[p_training])

                # Periodically update target net
                if (training_iter_id + 1) % self._args.ddqn_args.target_net_update_freq:
                    self._ray.wait([
                        self._ray.remote(la.update_target_net,
                                         p_training)
                        for la in self._la_handles
                    ])

                if (training_iter_id + 1) % SMOOTHING == 0:
                    print("RL-BR", "P" + str(p_training) + " iter", training_iter_id + 1)
                    accum_score /= SMOOTHING
                    logging_scores_each_p[p_training].append(accum_score)
                    logging_eps_per_p[p_training].append(
                        self._ray.get(self._ray.remote(self._ps_handle.get_eps, p_training)))
                    if p_training == 0:  # Only need to collect that once
                        logging_timesteps.append(training_iter_id + 1)
                    accum_score = 0.0

        # """""""""""""""""""
        # Logging
        # """""""""""""""""""
        if self._t_prof.log_verbose:
            logging_scores_avg = [
                sum([
                    logging_scores_each_p[i][t]
                    for i in range(self._eval_env_bldr.N_SEATS)
                ]) / self._eval_env_bldr.N_SEATS
                for t in range(len(logging_scores_each_p[0]))
            ]

            for i, logging_iter in enumerate(logging_timesteps):
                self._ray.remote(
                    self._chief_handle.add_scalar,
                    running_rew_exp, "RL-BR/Running Reward While Training", logging_iter,
                    logging_scores_avg[i])

                for p in range(self._eval_env_bldr.N_SEATS):
                    self._ray.remote(
                        self._chief_handle.add_scalar,
                        eps_exp[p], "RL-BR/Training Epsilon", logging_iter, logging_eps_per_p[p][i])

    def _get_gradients(self, p_id):
        grads = [
            self._ray.remote(la.get_grads,
                             p_id)
            for la in self._la_handles
        ]
        self._ray.wait(grads)

        return grads

    def _update_leaner_actors(self, update_eps_for_plyrs=None, update_net_for_plyrs=None):
        assert isinstance(update_net_for_plyrs, list) or update_net_for_plyrs is None
        assert isinstance(update_eps_for_plyrs, list) or update_eps_for_plyrs is None

        _update_net_per_p = [
            True if (update_net_for_plyrs is not None) and (p in update_net_for_plyrs) else False
            for p in range(self._t_prof.n_seats)
        ]

        _update_eps_per_p = [
            True if (update_eps_for_plyrs is not None) and (p in update_eps_for_plyrs) else False
            for p in range(self._t_prof.n_seats)
        ]

        la_batches = []
        n = len(self._la_handles)
        c = 0
        while n > c:
            s = min(n, c + self._t_prof.max_n_las_sync_simultaneously)
            la_batches.append(self._la_handles[c:s])
            if type(la_batches[-1]) is not list:
                la_batches[-1] = [la_batches[-1]]
            c = s

        eps = [None for _ in range(self._t_prof.n_seats)]
        nets = [None for _ in range(self._t_prof.n_seats)]
        for p_id in range(self._t_prof.n_seats):
            eps[p_id] = None if not _update_eps_per_p[p_id] else self._ray.remote(
                self._ps_handle.get_eps, p_id)

            nets[p_id] = None if not _update_net_per_p[p_id] else self._ray.remote(
                self._ps_handle.get_weights, p_id)

        for batch in la_batches:
            self._ray.wait([
                self._ray.remote(la.update,
                                 eps,
                                 nets)
                for la in batch
            ])

    @staticmethod
    def _compute_rlbr(n_hands_each_seat, rlbr_dqn_each_seat, rlbr_env_wrapper, opponent):
        agent_losses = np.empty(shape=n_hands_each_seat * 2, dtype=np.float32)

        for rlbr_seat_id in range(rlbr_env_wrapper.env.N_SEATS):

            rlbr_agent = rlbr_dqn_each_seat[rlbr_seat_id]

            for iteration_id in range(n_hands_each_seat):

                # """""""""""""""""
                # Reset
                # """""""""""""""""
                obs, r_for_all, done, info = _util.reset_episode_multi_action_space(rlbr_env_wrapper=rlbr_env_wrapper,
                                                                                    opponent_agent=opponent)
                range_idx_rlbr = rlbr_env_wrapper.env.get_range_idx(p_id=rlbr_seat_id)

                # """""""""""""""""
                # Play Episode
                # """""""""""""""""
                while not done:
                    p_id_acting = rlbr_env_wrapper.env.current_player.seat_id

                    # RL-BR acting
                    if p_id_acting == rlbr_seat_id:
                        action_int = rlbr_agent.select_br_a(
                            pub_obses=[obs],
                            range_idxs=np.array([range_idx_rlbr], dtype=np.int32),
                            legal_actions_lists=[rlbr_env_wrapper.env.get_legal_actions()],
                            explore=False,
                        )[0]
                        _util.notify_agent_multi_action_space(action_int=action_int, rlbr_seat_id=rlbr_seat_id,
                                                              rlbr_env_wrapper=rlbr_env_wrapper,
                                                              opponent_agent=opponent)

                        # Step
                        obs, r_for_all, done, info = rlbr_env_wrapper.step(action=action_int)

                    # EvalAgent (opponent) acting
                    else:
                        action_int, _ = opponent.get_action(step_env=True, need_probs=False)
                        obs, r_for_all, done, info = _util.step_from_opp_action(action_int=action_int,
                                                                                opponent=opponent,
                                                                                rlbr_env_wrapper=rlbr_env_wrapper)

                # add rews
                agent_losses[iteration_id + (rlbr_seat_id * n_hands_each_seat)] = r_for_all[rlbr_seat_id] \
                                                                                  * rlbr_env_wrapper.env.REWARD_SCALAR \
                                                                                  * rlbr_env_wrapper.env.EV_NORMALIZER

        return agent_losses



# Copyright (c) 2019 Eric Steinberger


import copy
import os
from os.path import join as ospj

import torch


class TrainingProfileBase:
    """
    A TrainingProfile should hold hyperparameters and an for a run of an algorithm.
    """

    def __init__(self,

                 # --- general
                 name,
                 log_verbose,
                 log_export_freq,
                 checkpoint_freq,
                 eval_agent_export_freq,

                 # --- env
                 game_cls,
                 env_bldr_cls,
                 start_chips,

                 # --- Evaluation
                 eval_modes_of_algo,
                 eval_stack_sizes,

                 module_args,

                 # --- Computing
                 path_data=None,
                 local_crayon_server_docker_address="localhost",
                 device_inference="cpu",
                 DISTRIBUTED=False,
                 CLUSTER=False,
                 DEBUGGING=False,

                 # --- Only relevant if running distributed
                 redis_head_adr=None,  # (str) IP under which the ray redis server can be reached
                 ):
        """
        Args:
            name (str):                             Under this name all logs, data, and checkpoints will appear.
            log_verbose (bool):                     Whether the program shall log detailed in Tensorboard.
            log_export_freq:                        Every X iterations, logs are pushed to the Crayon Docker container.
            checkpoint_freq (int):                  Every X iterations, make a recoverable copy of state of training.
            eval_agent_export_freq (int):        Every X iterations, an EvalAgent instance of the algo is exported.
            game_cls (PokerEnv subclass):           Class (not instance) to be trained in.
            env_bldr_cls (EnvBuilder subclass)      Class (not instance) to wrap the environment.
            start_chips (int):                      Standard stack size to initialize all players with.
            eval_modes_of_algo (tuple):             Tuple of algo-specific EvalAgent's eval modes
            eval_stack_sizes (tuple):               Tuple of lists of ints. if None, defaults to what's used in
                                                    training_profile.env_bldr.
            module_args (dict):                     All modules or parts of algorithms may have their own args. These
                                                    are stored in seperate objects and accessible under a certain string
                                                    key in the ""module_args"" dict.
            path_data:                              path to store data (e.g. checkpoints) the algorithm generates in.
                                                    If None, we will store data in a folder we create in your home dir.
            local_crayon_server_docker_address:     ip of crayon docker container (default: localhost)
            device_inference:                       "cpu" or "cuda". This device will be used for batched NN inference
            DISTRIBUTED (bool):                     Whether ray should be used at all.
            CLUSTER:                                requires "DISTRIBUTED==True".
                                                    If True, runs on many machines, if False, runs on local CPUs/GPUs.
            DEBUGGING (bool):                       Whether to use assert statements for debugging
            redis_head_adr:                         Only applicable if "CLUSTER==True". IP address under which the ray
                                                    head can be found.

        """

        # Assert basic modules were passed
        assert "env" in module_args

        # t_prof
        self.name = name
        self.log_verbose = log_verbose
        self.log_export_freq = log_export_freq
        self.checkpoint_freq = checkpoint_freq
        self.eval_agent_export_freq = eval_agent_export_freq

        self.module_args = module_args

        if CLUSTER:
            if redis_head_adr:
                self.redis_head_adr = redis_head_adr
            else:
                from ray import services

                self.redis_head_adr = services.get_node_ip_address() + ":6379"
        self.local_crayon_server_docker_address = local_crayon_server_docker_address

        self.DISTRIBUTED = DISTRIBUTED or CLUSTER
        self.CLUSTER = CLUSTER
        self.DEBUGGING = DEBUGGING
        self.HAVE_GPU = torch.cuda.is_available()

        self.n_seats = self.module_args["env"].n_seats

        # Eval
        self.eval_modes_of_algo = eval_modes_of_algo
        if eval_stack_sizes is None:
            if start_chips is None:
                self.eval_stack_sizes = [[game_cls.DEFAULT_STACK_SIZE for _ in range(self.n_seats)]]
            else:
                self.eval_stack_sizes = [copy.deepcopy(self.module_args["env"].starting_stack_sizes_list)]
        else:
            assert isinstance(eval_stack_sizes, tuple)
            assert isinstance(eval_stack_sizes[0], list)
            self.eval_stack_sizes = list(eval_stack_sizes)

        self.game_cls_str = game_cls.__name__
        self.env_builder_cls_str = env_bldr_cls.__name__

        assert isinstance(device_inference, str), "Please pass a string (either 'cpu' or 'cuda')!"
        self.device_inference = torch.device(device_inference)

        # Paths
        def get_root_path():
            return "C:\\" if os.name == 'nt' else os.path.expanduser('~/')

        self._data_path = path_data if path_data is not None else os.path.join(get_root_path(), "poker_ai_data")
        self.path_agent_export_storage = ospj(self._data_path, "eval_agent")
        self.path_log_storage = ospj(self._data_path, "logs")
        self.path_checkpoint = ospj(self._data_path, "checkpoint")
        self.path_trainingprofiles = ospj(self._data_path, "TrainingProfiles")

        for p in [self._data_path,
                  self.path_agent_export_storage,
                  self.path_log_storage,
                  self.path_checkpoint,
                  self.path_trainingprofiles,
                  ]:
            if (not os.path.exists(p)) and (not os.path.isfile(p)):
                os.makedirs(p)



# Copyright (c) 2019 Eric Steinberger


from os.path import join as ospj

from PokerRL.rl.MaybeRay import MaybeRay
from PokerRL.util import file_util


class WorkerBase:
    """
    The base-class for every worker that can run locally and on a cluster with the same code.
    """

    def __init__(self, t_prof):
        self._t_prof = t_prof

        if t_prof is None:
            self._ray = MaybeRay(runs_distributed=False, runs_cluster=False)
        else:
            self._ray = MaybeRay(runs_distributed=t_prof.DISTRIBUTED, runs_cluster=t_prof.CLUSTER)

    def checkpoint(self, curr_step):
        """
        Override if worker has state to store/load for checkpoints
        """
        pass

    def load_checkpoint(self, name_to_load, step):
        """
        Override if worker has state to store/load for checkpoints
        """
        pass

    def _get_checkpoint_file_path(self, name, step, cls, worker_id):
        path = ospj(self._t_prof.path_checkpoint, str(name), str(step))
        file_util.create_dir_if_not_exist(path)
        return ospj(path, cls.__name__ + "_" + str(worker_id) + ".pkl")



# Copyright (c) 2019 Eric Steinberger


"""
A collection of Poker games often used in computational poker research.
"""

from PokerRL.game.Poker import Poker
from PokerRL.game._.rl_env.game_rules import HoldemRules, LeducRules, FlopHoldemRules, BigLeducRules
from PokerRL.game._.rl_env.poker_types.DiscretizedPokerEnv import DiscretizedPokerEnv
from PokerRL.game._.rl_env.poker_types.LimitPokerEnv import LimitPokerEnv
from PokerRL.game._.rl_env.poker_types.NoLimitPokerEnv import NoLimitPokerEnv


# """""""""""""""
# Leduc Family
# """""""""""""""
class StandardLeduc(LeducRules, LimitPokerEnv):
    """
    Leduc Hold'em is a very small poker game meant for fast experimentation with new algorithms. It is played with 3
    ranks and 2 suits. Typically players place an ante of 1, the small_bet is 2, and the big_bet is 4. In this
    implementation we stick to 10/20/40 to keep order-of-magnitude consitency with the other games.
    """

    RULES = LeducRules
    IS_FIXED_LIMIT_GAME = True
    IS_POT_LIMIT_GAME = False
    MAX_N_RAISES_PER_ROUND = {
        Poker.PREFLOP: 2,
        Poker.FLOP: 2,
    }

    SMALL_BLIND = 0
    BIG_BLIND = 0
    ANTE = 1
    SMALL_BET = 2
    BIG_BET = 4
    DEFAULT_STACK_SIZE = 13

    EV_NORMALIZER = 1000.0 / ANTE  # Milli Antes
    WIN_METRIC = Poker.MeasureAnte

    ROUND_WHERE_BIG_BET_STARTS = Poker.FLOP

    def __init__(self, env_args, lut_holder, is_evaluating):
        LeducRules.__init__(self)
        LimitPokerEnv.__init__(self,
                               env_args=env_args,
                               lut_holder=lut_holder,
                               is_evaluating=is_evaluating)


class BigLeduc(BigLeducRules, LimitPokerEnv):
    """
    Leduc Hold'em is a very small poker game meant for fast experimentation with new algorithms. It is played with 3
    ranks and 2 suits. Typically players place an ante of 1, the small_bet is 2, and the big_bet is 4. In this
    implementation we stick to 10/20/40 to keep order-of-magnitude consitency with the other games.
    """

    RULES = BigLeducRules
    IS_FIXED_LIMIT_GAME = True
    IS_POT_LIMIT_GAME = False
    MAX_N_RAISES_PER_ROUND = {
        Poker.PREFLOP: 4,
        Poker.FLOP: 4,
    }

    SMALL_BLIND = 0
    BIG_BLIND = 0
    ANTE = 1
    SMALL_BET = 2
    BIG_BET = 4
    DEFAULT_STACK_SIZE = 25

    EV_NORMALIZER = 1000.0 / ANTE  # Milli Antes
    WIN_METRIC = Poker.MeasureAnte

    ROUND_WHERE_BIG_BET_STARTS = Poker.FLOP

    def __init__(self, env_args, lut_holder, is_evaluating):
        BigLeducRules.__init__(self)
        LimitPokerEnv.__init__(self,
                               env_args=env_args,
                               lut_holder=lut_holder,
                               is_evaluating=is_evaluating)


class NoLimitLeduc(LeducRules, NoLimitPokerEnv):
    """
    A variant of Leduc with no bet-cap in the no-limit format. It uses blinds instead of antes.
    """

    RULES = LeducRules
    IS_FIXED_LIMIT_GAME = False
    IS_POT_LIMIT_GAME = False

    SMALL_BLIND = 50
    BIG_BLIND = 100
    ANTE = 0
    DEFAULT_STACK_SIZE = 20000

    EV_NORMALIZER = 1000.0 / BIG_BLIND  # Milli BB
    WIN_METRIC = Poker.MeasureBB

    def __init__(self, env_args, lut_holder, is_evaluating):
        LeducRules.__init__(self)
        NoLimitPokerEnv.__init__(self,
                                 env_args=env_args,
                                 lut_holder=lut_holder,
                                 is_evaluating=is_evaluating)


class DiscretizedNLLeduc(LeducRules, DiscretizedPokerEnv):
    """
    Discretized version of No-Limit Leduc Hold'em (i.e. agents can only select from a predefined set of betsizes)
    """

    RULES = LeducRules
    IS_FIXED_LIMIT_GAME = False
    IS_POT_LIMIT_GAME = False

    SMALL_BLIND = 50
    BIG_BLIND = 100
    ANTE = 0
    DEFAULT_STACK_SIZE = 20000

    EV_NORMALIZER = 1000.0 / BIG_BLIND  # Milli BB
    WIN_METRIC = Poker.MeasureBB

    def __init__(self, env_args, lut_holder, is_evaluating):
        LeducRules.__init__(self)
        DiscretizedPokerEnv.__init__(self,
                                     env_args=env_args,
                                     lut_holder=lut_holder,
                                     is_evaluating=is_evaluating)


# """""""""""""""
# Hold'em Family
# """""""""""""""
class LimitHoldem(HoldemRules, LimitPokerEnv):
    """
    Fixed-Limit Texas Hold'em is a long-standing benchmark game that has been essentially solved by Bowling et al
    (http://science.sciencemag.org/content/347/6218/145) using an efficient distributed implementation of CFR+, an
    optimized version of regular CFR.
    """

    RULES = HoldemRules
    IS_FIXED_LIMIT_GAME = True
    IS_POT_LIMIT_GAME = False
    MAX_N_RAISES_PER_ROUND = {
        Poker.PREFLOP: 4,
        Poker.FLOP: 4,
        Poker.TURN: 4,
        Poker.RIVER: 4,
    }
    ROUND_WHERE_BIG_BET_STARTS = Poker.TURN

    SMALL_BLIND = 1
    BIG_BLIND = 2
    ANTE = 0
    SMALL_BET = 2
    BIG_BET = 4
    DEFAULT_STACK_SIZE = 48

    EV_NORMALIZER = 1000.0 / BIG_BLIND  # Milli BB
    WIN_METRIC = Poker.MeasureBB

    def __init__(self, env_args, lut_holder, is_evaluating):
        HoldemRules.__init__(self)
        LimitPokerEnv.__init__(self,
                               env_args=env_args,
                               lut_holder=lut_holder,
                               is_evaluating=is_evaluating)


class NoLimitHoldem(HoldemRules, NoLimitPokerEnv):
    """
    No-Limit Texas Hold'em is the largest poker game in which AI beat humans as of 31.08.2018. It has been the focus in
    work such as DeepStack (https://arxiv.org/abs/1701.01724) and Libratus
    (http://science.sciencemag.org/content/early/2017/12/15/science.aao1733).
    """

    RULES = HoldemRules
    IS_FIXED_LIMIT_GAME = False
    IS_POT_LIMIT_GAME = False

    SMALL_BLIND = 50
    BIG_BLIND = 100
    ANTE = 0
    DEFAULT_STACK_SIZE = 20000

    EV_NORMALIZER = 1000.0 / BIG_BLIND  # Milli BB
    WIN_METRIC = Poker.MeasureBB

    def __init__(self, env_args, lut_holder, is_evaluating):
        HoldemRules.__init__(self)
        NoLimitPokerEnv.__init__(self,
                                 env_args=env_args,
                                 lut_holder=lut_holder,
                                 is_evaluating=is_evaluating)


class DiscretizedNLHoldem(HoldemRules, DiscretizedPokerEnv):
    """
    Discretized version of No-Limit Texas Hold'em (i.e. agents can only select from a predefined set of betsizes)
    """

    RULES = HoldemRules
    IS_FIXED_LIMIT_GAME = False
    IS_POT_LIMIT_GAME = False

    SMALL_BLIND = 50
    BIG_BLIND = 100
    ANTE = 0
    DEFAULT_STACK_SIZE = 20000

    EV_NORMALIZER = 1000.0 / BIG_BLIND  # Milli BB
    WIN_METRIC = Poker.MeasureBB

    def __init__(self, env_args, lut_holder, is_evaluating):
        HoldemRules.__init__(self)
        DiscretizedPokerEnv.__init__(self,
                                     env_args=env_args,
                                     lut_holder=lut_holder,
                                     is_evaluating=is_evaluating)


class Flop5Holdem(FlopHoldemRules, LimitPokerEnv):
    RULES = FlopHoldemRules
    IS_FIXED_LIMIT_GAME = True
    IS_POT_LIMIT_GAME = False

    SMALL_BLIND = 50
    BIG_BLIND = 100
    ANTE = 0
    DEFAULT_STACK_SIZE = 3000

    EV_NORMALIZER = 1000.0 / BIG_BLIND  # Milli BB
    WIN_METRIC = Poker.MeasureBB

    MAX_N_RAISES_PER_ROUND = {
        Poker.PREFLOP: 2,  # is actually 1, but BB counts as a raise in this codebase
        Poker.FLOP: 2,
    }
    ROUND_WHERE_BIG_BET_STARTS = Poker.TURN

    UNITS_SMALL_BET = None
    UNITS_BIG_BET = None

    FIRST_ACTION_NO_CALL = True

    def __init__(self, env_args, lut_holder, is_evaluating):
        FlopHoldemRules.__init__(self)
        LimitPokerEnv.__init__(self,
                               env_args=env_args,
                               lut_holder=lut_holder,
                               is_evaluating=is_evaluating)

    def _adjust_raise(self, raise_total_amount_in_chips):
        return self.get_fraction_of_pot_raise(fraction=1.0, player_that_bets=self.current_player)


"""
register all new envs here!
"""
ALL_ENVS = [
    StandardLeduc,
    BigLeduc,
    NoLimitLeduc,
    DiscretizedNLLeduc,
    LimitHoldem,
    NoLimitHoldem,
    DiscretizedNLHoldem,
    Flop5Holdem,
]



from .MainPokerModuleFLAT import *
from .MainPokerModuleRNN import *
from .AvrgStrategyNet import *
from .AdvantageNet import *
from .DuelingQNet import *
from .NetWrapperBase import *
from .QNet import *



# Copyright (c) 2019 Eric Steinberger


"""
Some os-agnostic utility functions to read and write files easily.
"""
import json
import os
import pickle
from os.path import join as ospj


def create_dir_if_not_exist(path):
    if (not os.path.exists(path)) and (not os.path.isfile(path)):
        os.makedirs(path)


def get_all_files_in_dir(_dir):
    return [f for f in os.listdir(_dir) if os.path.isfile(os.path.join(_dir, f))]


def get_all_dirs_in_dir(_dir):
    return [d for d in os.listdir(_dir) if os.path.isdir(os.path.join(_dir, d))]


def get_file_name_without_ending_and_path_from_path(path):
    return os.path.splitext(os.path.basename(path))[0]


def write_dict_to_file_json(_dir, file_name, dictionary):
    create_dir_if_not_exist(_dir)
    with open(ospj(_dir, str(file_name) + ".json"), 'w') as file:
        file.write(json.dumps(dictionary))


def write_dict_to_file_js(_dir, file_name, dictionary):
    create_dir_if_not_exist(_dir)
    with open(ospj(_dir, str(file_name) + ".js"), 'w') as file:
        file.write("const data=" + json.dumps(dictionary))


def do_pickle(obj, path, file_name):
    create_dir_if_not_exist(path)
    with open(ospj(path, str(file_name) + ".pkl"), "wb") as pkl_file:
        pickle.dump(obj=obj, file=pkl_file, protocol=pickle.HIGHEST_PROTOCOL)


def load_pickle(path, file_name=None):
    if file_name is None:
        p = path
    else:
        p = ospj(path, str(file_name) + ".pkl")

    with open(p, "rb") as pkl_file:
        state = pickle.load(pkl_file)
    return state



# Copyright (c) 2019 Eric Steinberger
# Inspiration of architecture from DeepStack-Leduc (https://github.com/lifrordi/DeepStack-Leduc/tree/master/Source)


from PokerRL.game.PokerEnvStateDictEnums import EnvDictIdxs


class NodeBase:
    """Base node from which all nodes extend"""

    def __init__(self, env_state, tree, parent, p_id_acted_last, is_terminal, depth):
        """
        A NodeBase stores the STATE it had BEFORE someone acted in it. all actions trigger a child.
        The strategy in a node refers to the prob distr that the player that acts in the node
        acts with to PRODUCE THE NODE'S CHILDREN.
        """
        self.env_state = env_state
        self.parent = parent
        self.p_id_acting_next = self.env_state[EnvDictIdxs.current_player]
        self.p_id_acted_last = p_id_acted_last

        self.is_terminal = is_terminal
        self.depth = depth
        self.tree = tree

        # built in recursion
        self.allowed_actions = []
        self.children = []

        self.strategy = None  # p_id_acting_next' strategy: np.arr((range_size, n_actions), np.float32)
        self.reach_probs = None  # reach probs of all players: np.arr((n_seats, range_size), np.float32)
        self.ev = None  # EVs for all players: np.arr((n_seats, range_size), np.float32)
        self.ev_br = None  # EV vs BR:  np.arr((n_seats, range_size), np.float32)
        self.ev_weighted = None
        self.ev_br_weighted = None
        self.br_a_idx_in_child_arr_for_each_hand = None
        self.epsilon = None  # float
        self.exploitability = None  # float

        # Any algorithm using a PublicTree to run on (e.g. CFR, CFR+) can use this to hold a dict of vals
        self.data = None


class PlayerActionNode(NodeBase):

    def __init__(self, env_state, tree, parent, is_terminal, p_id_acted_last, action, depth, new_round_state=None):
        super().__init__(env_state=env_state, tree=tree, parent=parent, p_id_acted_last=p_id_acted_last,
                         is_terminal=is_terminal, depth=depth)
        if action is not None:
            self.action = action
        else:
            self.action = None
        self.new_round_state = new_round_state  # Hack to work with PokerEnv


class ChanceNode(NodeBase):

    def __init__(self, env_state, tree, parent, p_id_acted_last, is_terminal, depth):
        super().__init__(env_state=env_state, tree=tree, parent=parent, p_id_acted_last=p_id_acted_last,
                         is_terminal=is_terminal, depth=depth)

        self.action = "CHANCE"



# Copyright (c) 2019 Eric Steinberger


from PokerRL.rl.base_cls.workers.WorkerBase import WorkerBase


class ChiefBase(WorkerBase):
    """
    The Chief is the worker logs are sent to (although they are presented on the Driver worker). It is also responsible
    for tracking what iteration is currently running and tunneling up-to-date strategies between workers upon
    request. Furthermore, the Chief exports EvalAgents and logs to permanent storage.
    """

    def __init__(self, t_prof):
        """
        Args:
            t_prof (TrainingProfile)
        """
        super().__init__(t_prof=t_prof)
        self._experiment_names = {}
        self._log_buf = _LogBuffer()

    def pull_current_eval_strategy(self, last_iteration_receiver_has):
        """
        This function is used to update EvalAgents with the current NN weights / strategy / anything needed.
        The specific outputs of this function are completely dynamic, as they should be tunneled between
        algo-specific components.

        Args:
            last_iteration_receiver_has (int)

        Returns:
            Tuple(Weights, any_other_info)
        """
        raise NotImplementedError

    def export_agent(self, step):
        """
        Wraps the current strategy of the agent in an EvalAgent instance and exports that.
        """
        raise NotImplementedError

    # __________________________________________________ Logging API ___________________________________________________
    def create_experiment(self, name):
        """
        Registers a new experiment in the LogBuffer, which can later be seen and pulled into TensorBoard.

        Args:
            name (str): Name of the new experiment

        Returns:
            name (str): Name of the new experiment
        """
        return self._log_buf.create_experiment(name)

    def add_scalar(self, exp_name, graph_name, step, value):
        """
        Adds one datapoint to a 2D graph of an experiment in the LogBuffer.

        Args:
            exp_name (str): Name of the new experiment
            graph_name (str): Name of the graph into which to plot
            step (int): Timestep (x-axis) of the datapoint
            value (float): Value to plot at timestep ""step"".
        """
        self._log_buf.add_scalar(exp_name=exp_name, graph_name=graph_name, step=step, value=value)

    def get_new_values(self):
        return self._log_buf.get_new_values()


class _LogBuffer:
    """
    This class STORES logs. It does not write them to TensorBoard; for that use CrayonWrapper.
    """

    def __init__(self):
        self._experiments = {}
        self._new_values = {}

    def clear(self):
        self._experiments = {}

    def create_experiment(self, name):
        if name not in self._experiments.keys():
            self._experiments[name] = {}
        return name  # returning name for convenience when creating experiments in a list comp

    def add_scalar(self, exp_name, graph_name, step, value):
        if exp_name not in self._experiments.keys():
            raise AttributeError("Should create experiment before adding to it")

        if graph_name not in self._experiments[exp_name].keys():
            self._experiments[exp_name][graph_name] = []

        self._experiments[exp_name][graph_name].append([step, value])

        if exp_name not in self._new_values.keys():
            self._new_values[exp_name] = {}
        if graph_name not in self._new_values[exp_name].keys():
            self._new_values[exp_name][graph_name] = []

        self._new_values[exp_name][graph_name].append([step, value])

    def get_new_values(self):
        new_v = self._new_values
        self._new_values = {}
        experiment_names = list(self._experiments.keys())
        return new_v, experiment_names



# Copyright (c) 2019 Eric Steinberger


import torch

from PokerRL.rl.buffers._circular_base import CircularBufferBase


class CircularBufferFLAT(CircularBufferBase):

    def __init__(self, env_bldr, max_size):
        super().__init__(env_bldr=env_bldr, max_size=max_size)

        self.storage_device = torch.device("cpu")

        self._pub_obs_t_buffer = None
        self._action_t_buffer = None
        self._range_idx_buffer = None
        self._reward_buffer = None
        self._pub_obs_tp1_buffer = None
        self._legal_action_mask_t_buffer = None
        self._legal_action_mask_tp1_buffer = None
        self._done_buffer = None

        self.reset()

    @property
    def max_size(self):
        return self._max_size

    @property
    def size(self):
        return self._size

    def add_step(self, pub_obs_t, a_t, range_idx, r_t, legal_action_mask_t, pub_obs_tp1, done_tp1,
                 legal_action_mask_tp1):
        self._pub_obs_t_buffer[self._top] = torch.from_numpy(pub_obs_t).to(device=self.storage_device)
        self._action_t_buffer[self._top] = a_t
        self._range_idx_buffer[self._top] = range_idx
        self._reward_buffer[self._top] = r_t
        self._pub_obs_tp1_buffer[self._top] = torch.from_numpy(pub_obs_tp1).to(device=self.storage_device)
        self._legal_action_mask_tp1_buffer[self._top] = torch.from_numpy(legal_action_mask_tp1).to(
            device=self.storage_device)
        self._legal_action_mask_t_buffer[self._top] = torch.from_numpy(legal_action_mask_t).to(
            device=self.storage_device)
        self._done_buffer[self._top] = float(done_tp1)

        if self._size < self._max_size:
            self._size += 1

        self._top = (self._top + 1) % self._max_size

    def sample(self, device, batch_size):
        indices = torch.randint(0, self._size, (batch_size,), dtype=torch.long, device=device)

        return self._pub_obs_t_buffer[indices].to(device), \
               self._action_t_buffer[indices].to(device), \
               self._range_idx_buffer[indices].to(device), \
               self._legal_action_mask_t_buffer[indices].to(device), \
               self._reward_buffer[indices].to(device), \
               self._pub_obs_tp1_buffer[indices].to(device), \
               self._legal_action_mask_tp1_buffer[indices].to(device), \
               self._done_buffer[indices].to(device)

    def state_dict(self):
        return {
            "pub_obs_t_buffer": self._pub_obs_t_buffer.cpu().clone(),
            "action_t_buffer": self._action_t_buffer.cpu().clone(),
            "range_idx_buffer": self._range_idx_buffer.cpu().clone(),
            "reward_buffer": self._reward_buffer.cpu().clone(),
            "pub_obs_tp1_buffer": self._pub_obs_tp1_buffer.cpu().clone(),
            "legal_action_mask_t_buffer": self._legal_action_mask_t_buffer.cpu().clone(),
            "legal_action_mask_tp1_buffer": self._legal_action_mask_tp1_buffer.cpu().clone(),
            "done_buffer": self._done_buffer.cpu().clone(),
            "size": self._size,
            "top": self._top
        }

    def load_state_dict(self, state):
        self._pub_obs_t_buffer = state["pub_obs_t_buffer"]
        self._action_t_buffer = state["action_t_buffer"]
        self._range_idx_buffer = state["range_idx_buffer"]
        self._reward_buffer = state["reward_buffer"]
        self._pub_obs_tp1_buffer = state["pub_obs_tp1_buffer"]
        self._legal_action_mask_t_buffer = state["legal_action_mask_t_buffer"]
        self._legal_action_mask_tp1_buffer = state["legal_action_mask_tp1_buffer"]
        self._done_buffer = state["done_buffer"]
        self._size = state["size"]
        self._top = state["top"]

    def reset(self):
        super().reset()
        self._pub_obs_t_buffer = torch.empty(size=(self._max_size, self._env_bldr.pub_obs_size), dtype=torch.float32,
                                             device=self.storage_device)
        self._action_t_buffer = torch.empty(size=(self._max_size,), dtype=torch.long, device=self.storage_device)
        self._range_idx_buffer = torch.empty(size=(self._max_size,), dtype=torch.long, device=self.storage_device)
        self._reward_buffer = torch.empty(size=(self._max_size,), dtype=torch.float32, device=self.storage_device)
        self._pub_obs_tp1_buffer = torch.empty(size=(self._max_size, self._env_bldr.pub_obs_size),
                                               dtype=torch.float32, device=self.storage_device)
        self._legal_action_mask_t_buffer = torch.ByteTensor(size=(self._max_size, self._env_bldr.N_ACTIONS),
                                                            device=self.storage_device)
        self._legal_action_mask_tp1_buffer = torch.ByteTensor(size=(self._max_size, self._env_bldr.N_ACTIONS),
                                                              device=self.storage_device)
        self._done_buffer = torch.empty(size=(self._max_size,), dtype=torch.float32, device=self.storage_device)



# Copyright (c) 2019 Eric Steinberger


import numpy as np

from PokerRL.rl import rl_util
from PokerRL.rl.buffers._circular_base import BRMemorySaverBase


class _GameForBR:

    def __init__(self, range_idx):
        self._n_steps_in_game_memory = 0

        self._range_idx = range_idx

        self._obs_sequence = []

        self._obs_t_idxs_per_step = []
        self._obs_tp1_idxs_per_step = []

        self._action_buffer = []
        self._reward_buffer = []
        self._done_buffer = []
        self._legal_actions_mask_tp1_buffer = []
        self._legal_actions_mask_t_buffer = []

    @property
    def n_steps_in_game_memory(self):
        return self._n_steps_in_game_memory

    def add(self, o_t, a, legal_action_mask_t, rew, done, o_tp1, legal_action_mask_tp1):
        self._obs_t_idxs_per_step.append(o_t.shape[0])
        self._obs_tp1_idxs_per_step.append(o_tp1.shape[0])

        self._obs_sequence = np.copy(o_tp1)

        self._legal_actions_mask_tp1_buffer.append(np.copy(legal_action_mask_tp1))
        self._legal_actions_mask_t_buffer.append(np.copy(legal_action_mask_t))
        self._action_buffer.append(a)
        self._reward_buffer.append(rew)
        self._done_buffer.append(done)

        self._n_steps_in_game_memory += 1

    def sample(self):
        idx = np.random.randint(low=0, high=self._n_steps_in_game_memory)
        return {
            "o_t": self._obs_sequence[:self._obs_t_idxs_per_step[idx]],
            "o_tp1": self._obs_sequence[:self._obs_tp1_idxs_per_step[idx]],
            "mask_t": self._legal_actions_mask_t_buffer[idx],
            "mask_tp1": self._legal_actions_mask_tp1_buffer[idx],
            "a": self._action_buffer[idx],
            "rew": self._reward_buffer[idx],
            "done": self._done_buffer[idx],
            "range_idx": self._range_idx
        }


class BRMemorySaverRNN(BRMemorySaverBase):
    """ Interface for correct BR reward storing """

    def __init__(self, env_bldr, buffer):
        super().__init__(env_bldr=env_bldr, buffer=buffer)
        self._game_memory = None

    def add_terminal(self,
                     reward_p,
                     terminal_obs,
                     ):

        if self._intermediate_memory.is_level_1():
            self._game_memory.add(o_t=np.copy(self._intermediate_memory.obs_t),
                                  a=self._intermediate_memory.action,
                                  rew=reward_p,
                                  done=True,
                                  o_tp1=np.copy(terminal_obs),

                                  legal_action_mask_t=rl_util.get_legal_action_mask_np(
                                      n_actions=self._env_bldr.N_ACTIONS,
                                      legal_actions_list=self._intermediate_memory.legal_actions_list_t),

                                  legal_action_mask_tp1=rl_util.get_legal_action_mask_np(
                                      n_actions=self._env_bldr.N_ACTIONS,
                                      legal_actions_list=self._intermediate_memory.legal_actions_list_tp1)
                                  )

        if self._game_memory is not None:
            self._buffer.add_game(game=self._game_memory)
        self._intermediate_memory.reset()

    def add_non_terminal_experience(self,
                                    obs_t_before_acted,
                                    a_selected_t,
                                    legal_actions_list_t):

        self._intermediate_memory.add_experience(obs_t=obs_t_before_acted,
                                                 action_t=a_selected_t,
                                                 legal_actions_list_t=legal_actions_list_t)

        if self._intermediate_memory.is_complete():
            self._add_step_to_memory()
            self._intermediate_memory.step()

    def _add_step_to_memory(self):
        self._game_memory.add(o_t=np.copy(self._intermediate_memory.obs_t),
                              a=self._intermediate_memory.action,
                              rew=0.0,
                              done=False,
                              o_tp1=np.copy(self._intermediate_memory.obs_tp1),

                              legal_action_mask_t=rl_util.get_legal_action_mask_np(
                                  n_actions=self._env_bldr.N_ACTIONS,
                                  legal_actions_list=self._intermediate_memory.legal_actions_list_t),

                              legal_action_mask_tp1=rl_util.get_legal_action_mask_np(
                                  n_actions=self._env_bldr.N_ACTIONS,
                                  legal_actions_list=self._intermediate_memory.legal_actions_list_tp1))

    def reset(self, range_idx):
        """ Call with env reset """
        self._intermediate_memory.reset()
        self._game_memory = _GameForBR(range_idx=range_idx)


# Copyright (c) 2019 Eric Steinberger


from PokerRL.game._.wrappers._Wrapper import Wrapper


class VanillaWrapper(Wrapper):
    """
    This wrapper doesn't track any history in any way and always just shows the current observation. Thus, it does not
    have perfect recall.

    This wrapper is suitable for feedforward NN architectures.
    """

    def __init__(self, env, env_bldr_that_built_me):
        super().__init__(env=env, env_bldr_that_built_me=env_bldr_that_built_me)

    def _reset_state(self, **kwargs):
        pass

    def _pushback(self, env_obs):
        pass

    def print_obs(self, wrapped_obs):
        self.env.print_obs(wrapped_obs)

    def get_current_obs(self, env_obs=None):
        if env_obs:
            return env_obs
        return self.env.get_current_obs()

    def state_dict(self):
        return {
            "base": super().state_dict(),
        }

    def load_state_dict(self, state_dict):
        super().load_state_dict(state_dict=state_dict["base"])

    def set_to_public_tree_node_state(self, node):
        pass



# Copyright (c) 2019 Eric Steinberger


import time

import numpy as np

from PokerRL.game.Poker import Poker
from PokerRL.game._.rl_env.base.PokerEnv import PokerEnv as _PokerEnv
from PokerRL.game.poker_env_args import DiscretizedPokerEnvArgs


class DiscretizedPokerEnv(_PokerEnv):
    """
    To discretize No-Limit or Pot-Limit poker games, subclass this baseclass instaed of PokerEnv. It allows to define
    a set of bet_sizes (as fractions of the pot) that are then part of the action space. Contrary to the action format
    of PokerEnv tuple(action, raise_size), discretized envs have integer actions, where 0 is FOLD, 1 is CHECK/CALL and
    then come all specified raise sizes sorted ascending.
    """
    ARGS_CLS = DiscretizedPokerEnvArgs

    def __init__(self,
                 env_args,
                 lut_holder,
                 is_evaluating):

        """
        Args:
            env_args (DiscretePokerEnvArgs):    an instance of DiscretePokerEnvArgs, passing an instance of PokerEnvArgs
                                                will not work.
            is_evaluating (bool):               Whether the environment shall be spawned in evaluation mode
                                                (i.e. no randomization) or not.

        """
        assert isinstance(env_args, DiscretizedPokerEnvArgs)
        assert isinstance(env_args.bet_sizes_list_as_frac_of_pot, list)
        assert isinstance(env_args.uniform_action_interpolation, bool)
        super().__init__(env_args=env_args, lut_holder=lut_holder, is_evaluating=is_evaluating)

        self.bet_sizes_list_as_frac_of_pot = sorted(env_args.bet_sizes_list_as_frac_of_pot)  # ascending
        self.N_ACTIONS = env_args.N_ACTIONS
        self.uniform_action_interpolation = env_args.uniform_action_interpolation

    def _adjust_raise(self, raise_total_amount_in_chips):
        return max(self._get_current_total_min_raise(), raise_total_amount_in_chips)

    def _get_env_adjusted_action_formulation(self, action_int):
        """

        Args:
            action_int: integer representation of discretized action

        Returns:
            list: (action, raise_size) in "continuous" PokerEnv format

        """
        if action_int == 0:
            return [0, -1]
        if action_int == 1:
            return [1, -1]
        elif action_int > 1:
            selected = self.get_fraction_of_pot_raise(fraction=self.bet_sizes_list_as_frac_of_pot[action_int - 2],
                                                      player_that_bets=self.current_player)

            if self.uniform_action_interpolation and not self.IS_EVALUATING:
                # _________________________________________ The maximal amount _________________________________________
                if action_int == self.N_ACTIONS - 1:  # if highest bet in repertoire
                    if self.IS_POT_LIMIT_GAME:  # if PL game: max is pot
                        max_amnt = self.get_fraction_of_pot_raise(fraction=1.0, player_that_bets=self.current_player)
                    elif self.IS_FIXED_LIMIT_GAME:
                        raise EnvironmentError("Should not get here with a limit game!")
                    else:  # if NL game: max is allin
                        max_amnt = self.current_player.stack + self.current_player.current_bet

                else:  # else, max is the mean of the selected and the next-bigger bet size
                    bigger = self.get_fraction_of_pot_raise(fraction=self.bet_sizes_list_as_frac_of_pot[action_int - 1],
                                                            player_that_bets=self.current_player)
                    max_amnt = int(float(selected + bigger) / 2)

                # _________________________________________ The minimal amount _________________________________________
                if action_int == 2:  # if lowest bet in repertoire, min is minbet
                    min_amnt = self._get_current_total_min_raise()

                else:  # else, min is the mean of the selected and the next-smaller bet size
                    smaller = self.get_fraction_of_pot_raise(
                        fraction=self.bet_sizes_list_as_frac_of_pot[action_int - 3],
                        player_that_bets=self.current_player)
                    min_amnt = int(float(selected + smaller) / 2)

                if min_amnt >= max_amnt:  # can happen. The sampling would always be the same -> save time.
                    return [2, min_amnt]
                return [2, np.random.randint(low=min_amnt, high=max_amnt)]

            else:
                return [2, selected]
        else:
            raise ValueError(action_int)

    def get_legal_actions(self):
        """

        Returns:
            list:   a list with none, one or multiple actions of the set [FOLD, CHECK/CALL, BETSIZE_1, BETSIZE_2, ...]

        """
        legal_actions = []

        # Fold, Check/Call
        for a_int in [Poker.FOLD, Poker.CHECK_CALL]:
            _a = self._get_env_adjusted_action_formulation(action_int=a_int)
            if self._get_fixed_action(action=_a)[0] == a_int:
                legal_actions.append(a_int)

        # since raises are ascending in the list, we can simply loop and break
        _last_too_small = None
        for a in range(2, self.N_ACTIONS):  # only loops through raises
            adj_a = self._get_env_adjusted_action_formulation(action_int=a)
            fixed_a = self._get_fixed_action(action=adj_a)

            if adj_a[0] != fixed_a[0]:  # if we wanted to raise, but env told us we can not, dont append a raise
                break

            if (adj_a[1] < fixed_a[1]) and a < self.N_ACTIONS:
                _last_too_small = a
            else:
                if _last_too_small is not None:
                    legal_actions.append(_last_too_small)  # was to small, but rounded to minraise & is unique -> append
                    _last_too_small = None

                legal_actions.append(a)  # this action might be modified by env, but is legal and unique -> append

            if adj_a[1] > fixed_a[1]:  # if the raise was too big, an even bigger one will yield the same result
                break
        assert len(legal_actions) > 0
        return legal_actions

    def get_hand_rank(self, hand_2d, board_2d):
        """
        For docs, refer to PokerEnv.get_hand_rank_all_hands_on_given_boards()
        """
        raise NotImplementedError

    def get_hand_rank_all_hands_on_given_boards(self, boards_1d, lut_holder):
        """
        For docs, refer to PokerEnv.get_hand_rank_all_hands_on_given_boards()
        """
        raise NotImplementedError

    def get_random_action(self):
        legal = self.get_legal_actions()
        return legal[np.random.randint(len(legal))]

    def print_tutorial(self):
        print("____________________________________________ TUTORIAL ____________________________________________")
        print("Actions:")
        print("0 \tFold")
        print("1 \tCall")
        for i in range(2, self.N_ACTIONS):
            print(i, "\tRaise ", self.bet_sizes_list_as_frac_of_pot[i - 2] * 100, "% of the pot")

    def human_api_ask_action(self):
        """ Returns action in Tuple form. """
        while True:
            try:
                action_idx = int(
                    input("What action do you want to take as player " + str(self.current_player.seat_id) + "?"))
            except ValueError:
                print("You need to enter one of the allowed actions. Refer to the tutorial please.")
                continue
            if action_idx < Poker.FOLD or action_idx >= self.N_ACTIONS:
                print("Invalid action_idx! Please enter one of the allowed actions as described in the tutorial above.")
                continue
            break
        time.sleep(0.01)

        return self._get_fixed_action(self._get_env_adjusted_action_formulation(action_int=action_idx))



# Copyright (c) 2019 Eric Steinberger


import torch
import torch.nn as nn

from PokerRL.rl import rl_util


class MainPokerModuleRNN(nn.Module):
    """
    Feeds parts of the observation through different fc layers before the RNN

    Structure (each branch merge is a concat):

    Table & Player state --> FC -> ReLU -------------------------------------------.
    Board, private info  --> FC -> ReLU -> FC+Skip -> ReLU -> FC+Skip -> ReLU ---- cat --> FC -> ReLU -> RNN ->

    """

    def __init__(self,
                 env_bldr,
                 device,
                 mpm_args,
                 ):
        super().__init__()
        self.args = mpm_args

        self.env_bldr = env_bldr

        self.N_SEATS = self.env_bldr.N_SEATS
        self.device = device

        self.board_start = self.env_bldr.obs_board_idxs[0]
        self.board_len = len(self.env_bldr.obs_board_idxs)

        self.table_start = self.env_bldr.obs_table_state_idxs[0]
        self.table_len = len(self.env_bldr.obs_table_state_idxs)

        self.players_info_starts = [player_i_idxs[0] for player_i_idxs in self.env_bldr.obs_players_idxs]
        self.players_info_lens = [len(player_i_idxs) for player_i_idxs in self.env_bldr.obs_players_idxs]

        self.pub_obs_size = self.env_bldr.pub_obs_size
        self.priv_obs_size = self.env_bldr.priv_obs_size

        self._relu = nn.ReLU(inplace=False)

        if mpm_args.use_pre_layers:
            self.cards_fc_1 = nn.Linear(in_features=self.env_bldr.obs_size_board + self.env_bldr.priv_obs_size,
                                        out_features=mpm_args.n_cards_state_units)

            self.cards_fc_2 = nn.Linear(in_features=mpm_args.n_cards_state_units,
                                        out_features=mpm_args.n_cards_state_units)

            self.cards_fc_3 = nn.Linear(in_features=mpm_args.n_cards_state_units,
                                        out_features=mpm_args.n_cards_state_units)

            self.table_state_fc = nn.Linear(in_features=self.env_bldr.obs_size_table_state
                                                        + self.env_bldr.obs_size_player_info_each * self.N_SEATS,
                                            out_features=mpm_args.n_merge_and_table_layer_units)

            self.merge_fc = nn.Linear(in_features=mpm_args.n_cards_state_units + mpm_args.n_merge_and_table_layer_units,
                                      out_features=mpm_args.n_merge_and_table_layer_units)

            self.rnn = rl_util.str_to_rnn_cls(mpm_args.rnn_cls_str)(input_size=mpm_args.n_merge_and_table_layer_units,
                                                                    hidden_size=mpm_args.rnn_units,
                                                                    num_layers=mpm_args.rnn_stack,
                                                                    dropout=mpm_args.rnn_dropout,
                                                                    bidirectional=False,
                                                                    batch_first=False)

        else:
            """ Inputs all data directly into the rnn. """
            self.rnn = rl_util.str_to_rnn_cls(mpm_args.rnn_cls_str)(
                input_size=self.env_bldr.complete_obs_size,
                hidden_size=mpm_args.rnn_units,
                num_layers=mpm_args.rnn_stack,
                dropout=mpm_args.rnn_dropout,
                bidirectional=False,
                batch_first=False
            )
        self.lut_range_idx_2_priv_o = torch.from_numpy(self.env_bldr.lut_holder.LUT_RANGE_IDX_TO_PRIVATE_OBS)
        self.lut_range_idx_2_priv_o = self.lut_range_idx_2_priv_o.to(device=self.device, dtype=torch.float32)

        self.to(device)

    @property
    def output_units(self):
        return self.args.rnn_units

    def forward(self, pub_obses, range_idxs):
        """
        1. do list -> padded
        2. feed through pre-processing fc layers
        3. PackedSequence (sort, pack)
        4. rnn
        5. unpack (unpack re-sort)
        6. cut output to only last entry in sequence

        Args:
            pub_obses (list):                 list of np arrays of shape [np.arr([history_len, n_features]), ...)
            range_idxs (LongTensor):        range_idxs (one for each pub_obs) tensor([2, 421, 58, 912, ...])
        """

        # ____________________________________________ Packed Sequence _____________________________________________
        bs = len(pub_obses)

        # this is just for one history step! has shape [bs, n_priv_features] now
        priv_obs = self.lut_range_idx_2_priv_o[range_idxs]

        if bs > 1:

            seq_lens = torch.tensor([sample.shape[0] for sample in pub_obses], device=self.device, dtype=torch.int32)
            max_len = seq_lens.max().item()

            _pub_obs = pub_obses
            pub_obses = torch.zeros((max_len, bs, self.pub_obs_size), dtype=torch.float32, device=self.device)
            for i, pub in zip(range(bs), _pub_obs):
                pub_obses[:seq_lens[i], i] = torch.from_numpy(pub).to(self.device)

            # extend private obs to whole time series. Technically goes over the seq_len (to max len)
            priv_obs = priv_obs.unsqueeze(0).repeat(max_len, 1, 1)

            if self.args.use_pre_layers:
                # slice and feed through the fc layers. the result is a tensor: [max_len, bs, some_n_of_units]
                y = self._feed_through_pre_layers(pub_o=pub_obses, priv_o=priv_obs)
            else:
                # concat pub and priv obs
                y = torch.cat((priv_obs, pub_obses), dim=-1)

            # sort descending for packing
            seq_lens, idx_shifts = torch.sort(seq_lens, descending=True)
            y = y[:, idx_shifts, :]

            # pack
            y = torch.nn.utils.rnn.pack_padded_sequence(y, lengths=seq_lens, batch_first=False)

            # Feed through RNN
            y, _ = self.rnn(y)

            # Unpack Sequence
            y, seq_lens = nn.utils.rnn.pad_packed_sequence(y, batch_first=False)

            if self.args.sum_step_outputs:
                y = y.sum(0) * (1.0 / seq_lens.float()).unsqueeze(-1)
            else:  # use last element of each sequence in the batch, because we only need the last action
                y = y[seq_lens - 1, torch.arange(bs, device=self.device, dtype=torch.long), :].squeeze(dim=0)

            # sort the samples back to their original order
            idx_unsort_obs_t = torch.arange(bs, device=self.device, dtype=torch.long)
            idx_unsort_obs_t.scatter_(src=idx_unsort_obs_t.clone(), dim=0, index=idx_shifts)

            return y[idx_unsort_obs_t]

        else:
            seq_len = pub_obses[0].shape[0]
            pub_obses = torch.from_numpy(pub_obses[0]).to(self.device).view(seq_len, bs, self.pub_obs_size)
            priv_obs = priv_obs.unsqueeze(0).expand(seq_len, bs, self.priv_obs_size)

            if self.args.use_pre_layers:
                y = self._feed_through_pre_layers(pub_o=pub_obses, priv_o=priv_obs)
            else:
                # concat pub and priv obs
                y = torch.cat((priv_obs, pub_obses), dim=-1)

            y, _ = self.rnn(y)

            if self.args.sum_step_outputs:
                return y.sum(0) * (1.0 / seq_len)
            else:  # use last element of each sequence in the batch, because we only need the last action
                return y[seq_len - 1].view(bs, -1)

    def _feed_through_pre_layers(self, pub_o, priv_o):

        # """""""""""""""
        # Cards Body
        # """""""""""""""
        _cards_obs = torch.cat((priv_o, pub_o.narrow(dim=-1, start=self.board_start, length=self.board_len)), dim=-1)
        cards_out = self._relu(self.cards_fc_1(_cards_obs))
        cards_out = self._relu(self.cards_fc_2(cards_out) + cards_out)
        cards_out = self._relu(self.cards_fc_3(cards_out) + cards_out)

        # """""""""""""""
        # Table Body
        # """""""""""""""
        _table_obs = torch.cat(
            [
                pub_o.narrow(dim=-1, start=self.table_start, length=self.table_len)
            ]
            +
            [
                pub_o.narrow(dim=-1, start=self.players_info_starts[i],
                             length=self.players_info_lens[i])
                for i in range(self.N_SEATS)
            ]
            , dim=-1
        )
        table_out = self._relu(self.table_state_fc(_table_obs))

        # """""""""""""""
        # Merge Layer
        # """""""""""""""
        return self._relu(self.merge_fc(torch.cat([cards_out, table_out], dim=-1)))


class MPMArgsRNN:

    def __init__(self,
                 rnn_units,
                 rnn_stack,
                 rnn_dropout,
                 rnn_cls_str="lstm",
                 use_pre_layers=True,
                 n_cards_state_units=96,
                 n_merge_and_table_layer_units=32,
                 sum_step_outputs=False,
                 ):
        self.rnn_units = rnn_units
        self.rnn_stack = rnn_stack
        self.rnn_dropout = rnn_dropout
        self.rnn_cls_str = rnn_cls_str
        self.use_pre_layers = use_pre_layers
        self.n_cards_state_units = n_cards_state_units
        self.n_merge_and_table_layer_units = n_merge_and_table_layer_units
        self.sum_step_outputs = sum_step_outputs

    def get_mpm_cls(self):
        return MainPokerModuleRNN
